import pytest
from flask import Flask

from flarchitect.core.architect import Architect


def create_app(config: dict | None = None) -> Flask:
    app = Flask(__name__)
    if config:
        app.config.update(config)
    with app.app_context():
        Architect(app)

    @app.get("/ping")
    def ping() -> str:
        return "pong"

    return app


@pytest.fixture
def client_with_cors() -> Flask.test_client:  # type: ignore[name-defined]
    app = create_app(
        {
            "API_ENABLE_CORS": True,
            "CORS_RESOURCES": {r"/ping": {"origins": "*"}},
            "FULL_AUTO": False,
            "API_CREATE_DOCS": False,
        }
    )
    return app.test_client()


@pytest.fixture
def client_without_cors() -> Flask.test_client:  # type: ignore[name-defined]
    app = create_app(
        {
            "API_ENABLE_CORS": False,
            "FULL_AUTO": False,
            "API_CREATE_DOCS": False,
        }
    )
    return app.test_client()


def test_cors_header_present(client_with_cors) -> None:
    response = client_with_cors.get("/ping", headers={"Origin": "http://example.com"})
    assert response.status_code == 200
    assert response.headers.get("Access-Control-Allow-Origin") is not None


def test_cors_header_absent(client_without_cors) -> None:
    response = client_without_cors.get("/ping", headers={"Origin": "http://example.com"})
    assert response.status_code == 200
    assert "Access-Control-Allow-Origin" not in response.headers



import importlib.util
import os
import pprint
import re
import socket
from collections.abc import Callable
from importlib.machinery import ModuleSpec
from pathlib import Path
from typing import Any
from urllib.parse import urlparse
from xml.etree import ElementTree as ET

import inflect
from flask import Flask, current_app
from jinja2 import Environment, FileSystemLoader

from flarchitect.utils.config_helpers import get_config_or_model_meta
from flarchitect.utils.core_utils import convert_case, get_count

HTTP_METHODS = ["GET", "POST", "PATCH", "DELETE"]
DATE_FORMAT = "%Y-%m-%d"
DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"
p = inflect.engine()


def get_html_path() -> str:
    """Return the absolute path to the ``html`` templates directory.

    The path is resolved using the registered :class:`~flask.Flask` application
    when available. If the ``flarchitect`` extension is not registered or no
    application context is active, the path is determined by searching for an
    ``html`` folder relative to this file.

    Returns:
        str: Absolute path to the ``html`` directory. Returns an empty string if
        the directory cannot be located.
    """

    try:
        return current_app.extensions["flarchitect"].get_templates_path()  # type: ignore[index]
    except Exception:  # pragma: no cover - fallback when app context not available
        return find_html_directory() or ""


class AttributeInitializerMixin:
    """Used for the base Architect flask extension and others
    to initialise their config attributes."""

    def __init__(self, app: Flask, *args, **kwargs):
        self._set_class_attributes(**kwargs)
        self._set_app_config_attributes(app)
        super().__init__()

    def _set_app_config_attributes(self, app: Flask) -> None:
        """
        Sets class attributes from Flask app config if they exist.

        Args:
            app (Flask): The Flask application instance.
        """
        for key in vars(type(self)):
            if key.startswith("__"):
                continue
            config_key = key.upper().lstrip("_")
            if config_key in app.config:
                setattr(self, key, app.config[config_key])

    def _set_class_attributes(self, **kwargs: Any) -> None:
        """
        Sets class attributes from keyword arguments.

        Args:
            **kwargs: Keyword arguments representing class attributes.
        """
        for key in vars(type(self)):
            if key.startswith("__"):
                continue
            if key in kwargs:
                setattr(self, key, kwargs[key])


class AttributeInitialiserMixin(AttributeInitializerMixin):
    """UK spelling alias of :class:`AttributeInitializerMixin`."""

    pass


def find_html_directory(starting_directory: str | None = None) -> str | None:
    """Locate the nearest ``html`` directory by searching parent folders.

    Args:
        starting_directory: Directory to begin the search from. If ``None``,
            the search starts from the directory containing this module.

    Returns:
        str | None: Absolute path to the ``html`` directory or ``None`` if it
        cannot be found.
    """

    if starting_directory is None:
        starting_directory = os.path.abspath(os.path.dirname(__file__))

    contents = os.listdir(starting_directory)

    if "html" in contents and os.path.isdir(os.path.join(starting_directory, "html")):
        return os.path.join(starting_directory, "html")

    parent_directory = os.path.dirname(starting_directory)
    if starting_directory == parent_directory:
        return None

    return find_html_directory(parent_directory)


def manual_render_absolute_template(absolute_template_path: str, **kwargs: Any) -> str:
    """
    Manually renders a Jinja2 template given an absolute path.

    Args:
        absolute_template_path (str): The absolute path to the template.
        **kwargs: Additional keyword arguments to pass to the template.

    Returns:
        str: The rendered template as a string.
    """

    template_folder = os.path.join(get_html_path(), absolute_template_path)
    if template_folder.endswith(".html"):
        template_folder, template_filename = os.path.split(template_folder)

    env = Environment(loader=FileSystemLoader(template_folder))
    # ensure standard Flask globals like ``url_for`` are available
    try:
        from flask import url_for

        env.globals.update(url_for=url_for)
    except Exception:  # pragma: no cover - fallback when Flask not installed
        pass

    template = env.get_template(template_filename)
    return template.render(**kwargs)


def find_child_from_parent_dir(parent: str, child: str, current_dir: str = os.getcwd()) -> str | None:
    """
    Finds the directory of a child folder within a parent directory.

    Args:
        parent (str): The name of the parent directory.
        child (str): The name of the child directory.
        current_dir (str, optional): The current directory to start the search from.

    Returns:
        Optional[str]: The path to the child directory, or None if not found.
    """
    if os.path.basename(current_dir) == parent:
        for dirname in os.listdir(current_dir):
            if dirname == child:
                return os.path.join(current_dir, dirname)

    for dirname in os.listdir(current_dir):
        if dirname.startswith(".") or dirname == "node_modules":
            continue
        child_dir_path = os.path.join(current_dir, dirname)
        if os.path.isdir(child_dir_path):
            child_dir_path = find_child_from_parent_dir(parent, child, child_dir_path)
            if child_dir_path is not None:
                return child_dir_path

    return None


def check_rate_prerequisites(
    service: str,
    find_spec: Callable[[str], ModuleSpec | None] = importlib.util.find_spec,
) -> None:
    """Verify that dependencies for a cache backend are installed.

    Args:
        service: Name of the cache backend.
        find_spec: Function used to locate module specifications. Allows tests
            to supply a stub without monkeypatching.

    Raises:
        ImportError: If the required client library is missing.
    """
    back_end_spec = "or specify a cache service URI in the flask configuration with the key API_RATE_LIMIT_STORAGE_URI={URL}:{PORT}"
    if service == "Memcached":
        if find_spec("pymemcache") is None:
            raise ImportError("Memcached prerequisite not available. Please install pymemcache " + back_end_spec)
    elif service == "Redis":
        if find_spec("redis") is None:
            raise ImportError("Redis prerequisite not available. Please install redis-py " + back_end_spec)
    elif service == "MongoDB" and find_spec("pymongo") is None:
        raise ImportError("MongoDB prerequisite not available. Please install pymongo " + back_end_spec)


def check_rate_services(
    config_getter: Callable[[str, Any, Any], Any] = get_config_or_model_meta,
    prereq_checker: Callable[[str], None] = check_rate_prerequisites,
    socket_factory: Callable[..., socket.socket] = socket.socket,
) -> str | None:
    """Return the configured or automatically detected rate limit backend.

    Args:
        config_getter: Function used to obtain configuration values.
        prereq_checker: Callable used to validate backend prerequisites.
        socket_factory: Factory returning socket-like objects. Facilitates
            dependency injection for tests without monkeypatching.

    The function accepts explicitly configured cache URIs and validates that the
    scheme corresponds to a supported backend. ``memory://`` is permitted without
    a host component. If no configuration is provided, the function attempts to
    detect running local services for Memcached, Redis, or MongoDB and returns
    the appropriate URI.

    Returns:
        Optional[str]: The URI of the running service, or ``None`` if no service
        is found.
    """
    services = {
        "Memcached": 11211,
        "Redis": 6379,
        "MongoDB": 27017,
    }
    uri = config_getter("API_RATE_LIMIT_STORAGE_URI", default=None)
    if uri:
        parsed = urlparse(uri)
        scheme_map = {
            "memcached": "Memcached",
            "redis": "Redis",
            "mongodb": "MongoDB",
            "memory": None,
        }
        if not parsed.scheme:
            raise ValueError("Rate limit storage URI must include a scheme")
        if parsed.scheme not in scheme_map:
            raise ValueError(f"Unsupported rate limit storage backend: {parsed.scheme}")
        # In-memory backends do not require a network location. All others do.
        if parsed.scheme != "memory" and not parsed.netloc:
            raise ValueError("Rate limit storage URI must include a host")
        service_name = scheme_map[parsed.scheme]
        if service_name:
            prereq_checker(service_name)
        return uri

    # Allow disabling auto-detection in constrained environments (e.g. sandboxes/CI)
    if config_getter("API_RATE_LIMIT_AUTODETECT", default=True) is False:
        return None

    for service, port in services.items():
        try:
            sock = socket_factory(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(1)
            try:
                sock.connect(("127.0.0.1", port))
                sock.close()
                prereq_checker(service)

                rate_type = {
                    "Memcached": f"memcached://127.0.0.1:{port}",
                    "Redis": f"redis://127.0.0.1:{port}",
                    "MongoDB": f"mongodb://127.0.0.1:{port}",
                }
                return rate_type[service]

            except OSError:
                # Nothing listening on this port; try the next service
                continue
        except PermissionError:
            # Sockets disabled by the runtime (e.g. sandbox). Fall back to in-memory.
            return None

    return None


def validate_flask_limiter_rate_limit_string(rate_limit_str: str) -> bool:
    """
    Validates a rate limit string for Flask-Limiter.

    Args:
        rate_limit_str (str): The rate limit string to validate.

    Returns:
        bool: True if the rate limit string is valid, False otherwise.
    """
    pattern = re.compile(
        r"^\d+\s+per\s+(\d+\s+)?(second|minute|hour|day|seconds|minutes|hours|days)$",
        re.IGNORECASE,
    )
    return bool(pattern.match(rate_limit_str))


def search_all_keys(model: Any, key: str) -> bool:
    """Search for a specific key in all subclasses of a given model.

    Args:
        model (Any): The model class or instance to search in.
        key (str): The key to search for.

    Returns:
        bool: True if the key is found in any subclass, False otherwise.
    """
    return any(any(get_config_or_model_meta(key, model=subclass, method=method) for method in HTTP_METHODS) for subclass in model.__subclasses__())


def generate_readme_html(file_path: str | Path, *args: Any, **kwargs: Any) -> str:
    """Generate README content from a Jinja2 template.

    Args:
        file_path: Path to the Jinja2 template file. Relative paths are
            resolved from the project root.
        *args: Variable length argument list passed to ``render``.
        **kwargs: Arbitrary keyword arguments passed to ``render``.

    Returns:
        Rendered template content.

    Raises:
        FileNotFoundError: If ``file_path`` does not exist.
    """
    path = Path(file_path)
    if not path.is_absolute():
        # Resolve relative paths from the project root to avoid dependence on
        # the current working directory during test execution.
        path = Path(__file__).resolve().parents[2] / path

    if not path.exists():
        raise FileNotFoundError(f"{path} not found.")

    environment = Environment(loader=FileSystemLoader(str(path.parent)))
    template = environment.get_template(path.name)
    return template.render(*args, **kwargs)


def read_file_content(path: str) -> str:
    """Get the content of a file.

    Args:
        path (str): The path to the file.

    Returns:
        str: The content of the file.

    Raises:
        FileNotFoundError: If the file does not exist.
    """
    if os.path.exists(path):
        with open(path) as file:
            return file.read()
    raise FileNotFoundError(f"{path} not found.")


def case_no_change(s: str) -> str:
    """Return the input string unchanged.

    Args:
        s (str): The input string.

    Returns:
        str: The same input string.
    """
    return s


def pretty_print_dict(d: dict[Any, Any]) -> str:
    """Pretty print a dictionary.

    Args:
        d (Dict[Any, Any]): The dictionary to pretty print.

    Returns:
        str: The pretty-printed dictionary.
    """
    return pprint.pformat(d, indent=2)


def update_dict_if_flag_true(output: dict[str, Any], flag: bool, key: str, value: Any, case_func: Any) -> None:
    """Update a dictionary with a key-value pair if the flag is True.

    Args:
        output (Dict[str, Any]): The dictionary to update.
        flag (bool): The flag that controls whether to update.
        key (str): The key to update in the dictionary.
        value (Any): The value to associate with the key.
        case_func (Any): The function to convert the case of the key.
    """
    if flag:
        output.update({convert_case(key, case_func): value})


def make_base_dict() -> dict[str, Any]:
    """Create a base dictionary with configuration settings.

    Returns:
        Dict[str, Any]: The base dictionary with configuration settings.
    """
    output = {"value": "..."}
    # Default to 'snake' which is recognised by convert_case
    field_case = get_config_or_model_meta("API_FIELD_CASE", default="snake")

    config_options = [
        ("API_DUMP_DATETIME", "datetime", "2024-01-01T00:00:00.0000+00:00"),
        (
            "API_DUMP_VERSION",
            "api_version",
            get_config_or_model_meta("API_VERSION", default=True),
        ),
        ("API_DUMP_STATUS_CODE", "status_code", 200),
        ("API_DUMP_RESPONSE_MS", "response_ms", 15),
        ("API_DUMP_TOTAL_COUNT", "total_count", 10),
        ("API_DUMP_NULL_NEXT_URL", "next_url", "/api/example/url"),
        ("API_DUMP_NULL_PREVIOUS_URL", "previous_url", "null"),
        ("API_DUMP_NULL_ERRORS", "errors", "null", False),
    ]

    for config, key, value, *defaults in config_options:
        flag = get_config_or_model_meta(config, default=defaults[0] if defaults else True)
        update_dict_if_flag_true(output, flag, key, value, field_case)

    return output


def pluralize_last_word(converted_name: str) -> str:
    """
    Pluralize the last word of the converted name
    while preserving the rest of the name and its case.

    Args:
        converted_name (str): The name after case conversion.

    Returns:
        str: The name with the last word pluralized.
    """
    delimiters = {"_": "snake", "-": "kebab"}
    delimiter = next((d for d in delimiters if d in converted_name), "")

    words = converted_name.split(delimiter) if delimiter else re.findall(r"[A-Z]?[a-z]+|[A-Z]+(?![a-z])", converted_name)
    last_word = words[-1]
    last_word_pluralized = p.plural(p.singular_noun(last_word) or last_word)

    words[-1] = last_word_pluralized
    pluralized_name = delimiter.join(words)

    if delimiters.get(delimiter) in ["screaming_snake", "screaming_kebab"]:
        pluralized_name = pluralized_name.upper()

    return pluralized_name


def normalize_key(key: str) -> str:
    """
    Converts a key to uppercase.

    Args:
        key (str): The key to be normalized.

    Returns:
        str: The normalized key.
    """
    return key.upper()


def xml_to_dict(xml_data: str | bytes) -> dict[str, Any]:
    """
    Converts an XML string or bytes into a dictionary.
    Args:
        xml_data (Union[str, bytes]): The XML data.
    Returns:
        Dict[str, Any]: The resulting dictionary.
    """
    xml_data = xml_data.decode() if hasattr(xml_data, "decode") else xml_data

    def element_to_dict(element: ET.Element) -> Any:
        if not list(element) and (element.text is None or not element.text.strip()):
            return None
        if element.text and element.text.strip() and not list(element):
            return element.text.strip()
        result = {}
        for child in element:
            child_result = element_to_dict(child)
            if child.tag not in result:
                result[child.tag] = child_result
            else:
                if not isinstance(result[child.tag], list):
                    result[child.tag] = [result[child.tag]]
                result[child.tag].append(child_result)
        return result

    try:
        root = ET.fromstring(xml_data)
    except ET.ParseError as e:
        raise ValueError("Invalid XML data provided") from e
    return {root.tag: element_to_dict(root)}


def handle_result(result: Any) -> tuple[int, Any, int, str | None, str | None]:
    """Normalise values returned from route handlers.

    Args:
        result: The value returned by a view function. It may be a
            raw value, a ``(value, status_code)`` tuple or a dictionary
            containing ``query`` and pagination metadata.

    Returns:
        tuple[int, Any, int, str | None, str | None]:
            A tuple of ``(status_code, value, count, next_url, previous_url)``
            ready for ``create_response``.
    """

    status_code, value, count, next_url, previous_url = HTTP_OK, result, 1, None, None

    if isinstance(result, tuple) and len(result) == 2 and isinstance(result[1], int):
        result, status_code = result
    if isinstance(result, dict):
        value = result.get("query", result)
        count = get_count(result, value)
        next_url = result.get("next_url")
        previous_url = result.get("previous_url")
    else:
        value = result

    return status_code, value, count, next_url, previous_url


HTTP_OK = 200
HTTP_BAD_REQUEST = 400
HTTP_INTERNAL_SERVER_ERROR = 500
HTTP_UNAUTHORIZED = 401
HTTP_FORBIDDEN = 403



from flask import Flask

from demo.basic_factory.basic_factory.config import Config
from demo.basic_factory.basic_factory.extensions import db, schema
from demo.utils.helpers import load_dummy_database


def create_app(config: dict = None):
    """
    Creates the flask app.
    Args:
        config (Optional[dict]): The configuration dictionary.

    Returns:

    """
    app = Flask(__name__)
    app.config.from_object(Config)
    if config:
        app.config.update(config)

    db.init_app(app)

    with app.app_context():
        # Import models for their side effects so SQLAlchemy can register them
        import demo.basic_factory.basic_factory.models  # noqa: F401

        db.create_all()
        load_dummy_database(db)
        schema.init_app(app)

    return app



"""Tests for tag registration in OpenAPI specs."""

from demo.basic_factory.basic_factory import create_app


def test_spec_contains_tags() -> None:
    """Ensure that generated specs include root-level tag definitions."""
    app = create_app()
    client = app.test_client()
    spec = client.get("/openapi.json").get_json()
    assert "tags" in spec
    tag_names = {t["name"] for t in spec["tags"]}
    assert {"Books", "Categories"}.issubset(tag_names)



"""Unit tests for :class:`Architect` helper methods."""

from types import SimpleNamespace

import pytest
from flask import Flask
from marshmallow import Schema

from flarchitect import Architect
from flarchitect.exceptions import CustomHTTPException


class DummySchema(Schema):
    """Placeholder schema for decorator tests."""


def create_app(**config) -> Flask:
    app = Flask(__name__)
    app.config.update(
        FULL_AUTO=False,
        API_CREATE_DOCS=False,
        API_RATE_LIMIT_STORAGE_URI="memory://",
        **config,
    )
    return app


def test_handle_auth_custom(monkeypatch):
    """Verify ``_handle_auth`` respects custom authentication."""

    app = create_app(API_AUTHENTICATE_METHOD=["custom"], API_CUSTOM_AUTH=lambda: True)
    with app.app_context():
        architect = Architect(app=app)
    with app.test_request_context("/"):
        architect._handle_auth(model=None, output_schema=None, input_schema=None, auth_flag=True)

    app.config["API_CUSTOM_AUTH"] = lambda: False
    with app.test_request_context("/"), pytest.raises(CustomHTTPException):
        architect._handle_auth(model=None, output_schema=None, input_schema=None, auth_flag=True)


def test_apply_schemas_uses_handle_one(monkeypatch):
    """``_apply_schemas`` should delegate to ``handle_one`` when ``many`` is ``False``."""

    app = create_app()
    with app.app_context():
        architect = Architect(app=app)
    called = {"one": False, "many": False}

    def fake_handle_one(*_args, **_kwargs):  # type: ignore[unused-ignore]
        called["one"] = True

        def decorator(f):
            return f

        return decorator

    def fake_handle_many(*_args, **_kwargs):  # type: ignore[unused-ignore]
        called["many"] = True

        def decorator(f):
            return f

        return decorator

    monkeypatch.setattr("flarchitect.core.architect.handle_one", fake_handle_one)
    monkeypatch.setattr("flarchitect.core.architect.handle_many", fake_handle_many)

    def dummy():
        return "ok"

    architect._apply_schemas(dummy, DummySchema, None, many=False)
    assert called["one"] and not called["many"]


def test_apply_schemas_uses_handle_many(monkeypatch):
    """``_apply_schemas`` should delegate to ``handle_many`` when ``many`` is ``True``."""

    app = create_app()
    with app.app_context():
        architect = Architect(app=app)
    called = {"one": False, "many": False}

    def fake_handle_one(*_args, **_kwargs):  # type: ignore[unused-ignore]
        called["one"] = True

        def decorator(f):
            return f

        return decorator

    def fake_handle_many(*_args, **_kwargs):  # type: ignore[unused-ignore]
        called["many"] = True

        def decorator(f):
            return f

        return decorator

    monkeypatch.setattr("flarchitect.core.architect.handle_one", fake_handle_one)
    monkeypatch.setattr("flarchitect.core.architect.handle_many", fake_handle_many)

    def dummy():
        return "ok"

    architect._apply_schemas(dummy, DummySchema, None, many=True)
    assert called["many"] and not called["one"]


def test_apply_rate_limit_valid(monkeypatch):
    """A valid rate limit string should invoke the limiter with the value."""

    app = create_app(API_RATE_LIMIT="1 per minute")
    recorded = {}

    def fake_limit(rate):
        recorded["rate"] = rate

        def decorator(f):
            return f

        return decorator

    with app.app_context():
        architect = Architect(app=app)
        architect.limiter.limit = fake_limit  # type: ignore[method-assign]

        def dummy():
            return "ok"

        architect._apply_rate_limit(dummy, model=None, output_schema=None, input_schema=None)

    assert recorded["rate"] == "1 per minute"


def test_apply_rate_limit_invalid_logs(monkeypatch):
    """An invalid rate limit should log an error and return the original function."""

    app = create_app(API_RATE_LIMIT="invalid")
    with app.app_context():
        architect = Architect(app=app)

    monkeypatch.setattr(
        "flarchitect.core.architect.validate_flask_limiter_rate_limit_string",
        lambda _s: False,
    )

    monkeypatch.setattr(
        "flarchitect.core.architect.find_rule_by_function",
        lambda _self, _f: SimpleNamespace(rule="/dummy"),
    )

    messages = {}

    def fake_error(msg):
        messages["msg"] = msg

    monkeypatch.setattr("flarchitect.core.architect.logger.error", fake_error)

    def dummy():
        return "ok"

    with app.app_context():
        result = architect._apply_rate_limit(dummy, model=None, output_schema=None, input_schema=None)

    assert messages["msg"].startswith("Rate limit definition not a string")
    assert result is dummy



"""Configuration module for the scaffolding application."""

from __future__ import annotations

import os

from .extensions import db
from .models import User


class Config:
    """Base configuration with sensible defaults.

    Environment variables may override these values. Each option lists
    possible values and defaults for clarity.
    """

    # Flask settings
    SECRET_KEY = os.getenv("SECRET_KEY", "dev")
    SQLALCHEMY_DATABASE_URI = os.getenv("DATABASE_URL", "sqlite:///:memory:")
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    DEBUG = os.getenv("DEBUG", False)

    # flarchitect configuration
    API_BASE_MODEL = db.Model
    API_TITLE = "Scaffolding API"
    API_VERSION = "0.1.0"
    API_VERBOSITY_LEVEL = 4
    API_ALLOW_NESTED_WRITES = True
    API_CREATE_DOCS = True
    API_ENABLE_CORS = False
    API_DOCUMENTATION_PASSWORD = os.getenv("API_DOCUMENTATION_PASSWORD")
    API_DOCUMENTATION_REQUIRE_AUTH = False
    API_SOFT_DELETE = True
    API_SOFT_DELETE_ATTRIBUTE = "deleted"
    API_SOFT_DELETE_VALUES = (False, True)

    # Authentication configuration
    API_AUTHENTICATE = True
    API_AUTHENTICATE_METHOD = ["jwt"]
    API_USER_MODEL = User
    API_USER_LOOKUP_FIELD = "username"
    API_CREDENTIAL_CHECK_METHOD = "check_password"
    ACCESS_SECRET_KEY = os.getenv("ACCESS_SECRET_KEY", "access-secret")
    REFRESH_SECRET_KEY = os.getenv("REFRESH_SECRET_KEY", "refresh-secret")
    API_JWT_EXPIRY_TIME = 15
    API_JWT_REFRESH_EXPIRY_TIME = 1440
    API_DOCS_STYLE = "redoc"
    # Demo-specific options
    USERNAME_MIN_LENGTH = 3


class TestingConfig(Config):
    """Configuration used during unit tests."""

    TESTING = True
    SQLALCHEMY_DATABASE_URI = "sqlite:///:memory:"
    API_JWT_EXPIRY_TIME = 1
    API_JWT_REFRESH_EXPIRY_TIME = 2



from demo.basic_factory.basic_factory import create_app


def test_default_spec_route():
    """The OpenAPI spec is served at the default path."""
    app = create_app()
    client = app.test_client()
    resp = client.get("/openapi.json")
    assert resp.status_code == 200
    data = resp.get_json()
    assert isinstance(data, dict)
    assert data.get("openapi")
    assert data["components"]["securitySchemes"]["bearerAuth"] == {
        "type": "http",
        "scheme": "bearer",
        "bearerFormat": "JWT",
    }


def test_custom_spec_route():
    """The spec route can be overridden via configuration."""
    app = create_app({"API_SPEC_ROUTE": "/spec.json"})
    client = app.test_client()
    resp = client.get("/spec.json")
    assert resp.status_code == 200
    assert resp.get_json().get("openapi")



"""Tests for response filtering utilities."""

from __future__ import annotations

from typing import Any

import pytest

from flarchitect.utils.response_filters import _filter_response_data


@pytest.mark.parametrize(
    "config, expected_keys",
    [
        (
            {
                "API_DUMP_DATETIME": True,
                "API_DUMP_VERSION": True,
                "API_DUMP_STATUS_CODE": True,
                "API_DUMP_RESPONSE_MS": True,
                "API_DUMP_TOTAL_COUNT": True,
                "API_DUMP_NULL_NEXT_URL": True,
                "API_DUMP_NULL_PREVIOUS_URL": True,
                "API_DUMP_NULL_ERRORS": True,
            },
            {
                "datetime",
                "api_version",
                "status_code",
                "response_ms",
                "total_count",
                "next_url",
                "previous_url",
                "errors",
                "payload",
            },
        ),
        (
            {
                "API_DUMP_DATETIME": False,
                "API_DUMP_VERSION": False,
                "API_DUMP_STATUS_CODE": False,
                "API_DUMP_RESPONSE_MS": False,
                "API_DUMP_TOTAL_COUNT": False,
                "API_DUMP_NULL_NEXT_URL": False,
                "API_DUMP_NULL_PREVIOUS_URL": False,
                "API_DUMP_NULL_ERRORS": False,
            },
            {"payload"},
        ),
        (
            {
                "API_DUMP_DATETIME": True,
                "API_DUMP_VERSION": False,
                "API_DUMP_STATUS_CODE": True,
                "API_DUMP_RESPONSE_MS": False,
                "API_DUMP_TOTAL_COUNT": True,
                "API_DUMP_NULL_NEXT_URL": True,
                "API_DUMP_NULL_PREVIOUS_URL": False,
                "API_DUMP_NULL_ERRORS": True,
            },
            {"datetime", "status_code", "total_count", "next_url", "errors", "payload"},
        ),
    ],
)
def test_filter_response_data_removes_and_keeps_keys(
    monkeypatch: pytest.MonkeyPatch, config: dict[str, bool], expected_keys: set[str]
) -> None:
    """Ensure ``_filter_response_data`` honors configuration switches.

    Args:
        monkeypatch: Pytest fixture for modifying objects during the test.
        config: Mapping of configuration flags to their desired values.
        expected_keys: Keys that should remain after filtering.
    """
    data: dict[str, Any] = {
        "datetime": "2024-01-01T00:00:00+00:00",
        "api_version": "v1",
        "status_code": 200,
        "response_ms": 10,
        "total_count": 3,
        "next_url": None,
        "previous_url": "",
        "errors": [],
        "payload": {"id": 1},
    }

    def fake_get_config_or_model_meta(
        key: str, *_, default: Any | None = None, **__
    ) -> Any:
        """Return configuration override for the given key."""
        return config.get(key, default)

    monkeypatch.setattr(
        "flarchitect.utils.response_filters.get_config_or_model_meta",
        fake_get_config_or_model_meta,
    )

    filtered = _filter_response_data(data.copy())
    expected = {k: data[k] for k in expected_keys}

    assert filtered == expected



import pytest

from demo.basic_factory.basic_factory import create_app
from demo.basic_factory.basic_factory.extensions import db
from demo.basic_factory.basic_factory.models import Book
from flarchitect.database.operations import paginate_query


@pytest.fixture
def app():
    """Create a demo application for testing."""
    return create_app()


def test_paginate_query_returns_paginated_query_and_default(app):
    """Paginate a query and return default pagination size."""
    with app.app_context():
        query = db.session.query(Book)
        paginated, default_size = paginate_query(query, page=1, items_per_page=1)
        assert len(paginated.items) == 1
        assert default_size == 20



"""Database models for the scaffolding application with advanced options."""

from __future__ import annotations

from flask import current_app
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from validators import email as validate_email
from werkzeug.security import check_password_hash, generate_password_hash

from .extensions import db


def _titleize_name(obj: Item) -> Item:
    """Ensure item names are title-cased before saving."""

    obj.name = obj.name.title()
    return obj


class Category(db.Model):
    """Grouping for items to demonstrate nested writes."""

    __tablename__ = "category"

    class Meta:
        tag_group = "Core"
        tag = "Categories"
        allow_nested_writes = True
        description = {
            "GET": "Retrieve categories and their items.",
            "POST": "Create a new category.",
        }

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(db.String(80), unique=True, nullable=False, info={"description": "Category name"})

    items: Mapped[list[Item]] = relationship(back_populates="category")

    @validates("name")
    def validate_name(self, _key: str, value: str) -> str:
        """Ensure category names are not empty."""

        if not value.strip():
            raise ValueError("Category name cannot be empty")
        return value.strip()


class User(db.Model):
    """User account model with validation and password helpers."""

    __tablename__ = "user"

    class Meta:
        tag_group = "Core"
        tag = "Users"
        allow_nested_writes = False
        allowed_methods = ["GET", "POST"]
        post_rate_limit = "5 per minute"
        description = {
            "GET": "Retrieve users.",
            "POST": "Create a new user.",
        }

    id: Mapped[int] = mapped_column(primary_key=True)
    username: Mapped[str] = mapped_column(db.String(80), unique=True, nullable=False, info={"description": "Unique username"})
    email: Mapped[str] = mapped_column(
        db.String(120),
        unique=True,
        nullable=False,
        info={
            "description": "User email address.",
            "format": "email",
            "validator": "email",
            "validator_message": "Invalid email address.",
        },
    )
    website: Mapped[str | None] = mapped_column(db.String(255), info={"description": "User website", "format": "uri"})
    password_hash: Mapped[str] = mapped_column(db.String(128), nullable=False)

    items: Mapped[list[Item]] = relationship(back_populates="owner")

    @validates("email")
    def validate_email(self, _key: str, address: str) -> str:
        """Ensure e-mail addresses are valid."""

        if not validate_email(address):
            raise ValueError("Invalid email address")
        return address

    @validates("username")
    def validate_username(self, _key: str, username: str) -> str:
        """Validate username length from configuration."""

        min_len = int(current_app.config.get("USERNAME_MIN_LENGTH", 3))
        if len(username) < min_len:
            raise ValueError(f"Username must be at least {min_len} characters long")
        return username

    def set_password(self, password: str) -> None:
        """Hash and store the user's password."""

        self.password_hash = generate_password_hash(password)

    def check_password(self, password: str) -> bool:
        """Check a password against the stored hash."""

        return check_password_hash(self.password_hash, password)


class Item(db.Model):
    """Example model representing a user-owned item."""

    __tablename__ = "item"

    class Meta:
        tag_group = "Core"
        tag = "Items"
        allow_nested_writes = True
        add_callback = staticmethod(lambda obj, model: _titleize_name(obj))
        description = {
            "GET": "Retrieve items with their owners and categories.",
            "POST": "Create an item, optionally with its category.",
            "PATCH": "Update an item's details.",
        }
        patch_rate_limit = "10 per minute"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(db.String(120), nullable=False, info={"description": "Item name"})
    price: Mapped[float | None] = mapped_column(db.Numeric(10, 2), nullable=True, info={"description": "Item price", "example": 9.99})
    owner_id: Mapped[int] = mapped_column(db.ForeignKey("user.id"), nullable=False)
    owner: Mapped[User] = relationship(back_populates="items")
    category_id: Mapped[int | None] = mapped_column(db.ForeignKey("category.id"), nullable=True)
    category: Mapped[Category | None] = relationship(back_populates="items")

    @validates("price")
    def validate_price(self, _key: str, value: float | None) -> float | None:
        """Ensure prices are positive."""

        if value is not None and value < 0:
            raise ValueError("Price must be positive")
        return value



"""Tests for refresh token rotation, revocation (deny-list), and auditing."""

from __future__ import annotations

import datetime
from collections.abc import Generator

import pytest
from flask import Flask
from flask.testing import FlaskClient
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy.pool import StaticPool

from flarchitect import Architect
from flarchitect.authentication.token_store import RefreshToken, get_session


db = SQLAlchemy()


class User(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String, unique=True)
    password = db.Column(db.String)

    def check_password(self, password: str) -> bool:  # pragma: no cover - simple helper
        return self.password == password


@pytest.fixture()
def client_jwt_app() -> Generator[FlaskClient, None, None]:
    app = Flask(__name__)
    app.config.update(
        SQLALCHEMY_DATABASE_URI="sqlite:///:memory:",
        SQLALCHEMY_ENGINE_OPTIONS={"poolclass": StaticPool},
        SQLALCHEMY_TRACK_MODIFICATIONS=False,
        FULL_AUTO=False,
        API_CREATE_DOCS=False,
        API_RATE_LIMIT_STORAGE_URI="memory://",
        API_AUTHENTICATE_METHOD=["jwt"],
        API_USER_MODEL=User,
        API_USER_LOOKUP_FIELD="username",
        API_CREDENTIAL_CHECK_METHOD="check_password",
    )
    app.config["ACCESS_SECRET_KEY"] = "access"
    app.config["REFRESH_SECRET_KEY"] = "refresh"
    db.init_app(app)

    with app.app_context():
        architect = Architect(app=app)
        architect.init_api(app=app, api_full_auto=False)
        architect.api.make_auth_routes()

        db.create_all()
        user = User(username="eve", password="secret")
        db.session.add(user)
        db.session.commit()

        yield app.test_client()


def test_refresh_rotation_marks_old_revoked_and_links_new(client_jwt_app: FlaskClient) -> None:
    # Login to obtain tokens
    resp = client_jwt_app.post("/auth/login", json={"username": "eve", "password": "secret"})
    assert resp.status_code == 200
    data = resp.get_json()["value"]
    old_refresh = data["refresh_token"]

    # Refresh to rotate
    resp2 = client_jwt_app.post("/auth/refresh", json={"refresh_token": old_refresh})
    assert resp2.status_code == 200
    data2 = resp2.get_json()["value"]
    new_refresh = data2["refresh_token"]

    # Old token should be revoked and linked to new token, with last_used_at set
    with get_session(RefreshToken) as session:
        row = session.get(RefreshToken, old_refresh)
        assert row is not None
        assert row.revoked is True
        assert row.replaced_by == new_refresh
        assert row.last_used_at is not None

    # Using the old refresh again should fail (hidden by get_refresh_token logic and route validation)
    resp3 = client_jwt_app.post("/auth/refresh", json={"refresh_token": old_refresh})
    assert resp3.status_code in (401, 403)


def test_revoke_refresh_token_blocks_use(client_jwt_app: FlaskClient) -> None:
    # Login to obtain tokens
    resp = client_jwt_app.post("/auth/login", json={"username": "eve", "password": "secret"})
    assert resp.status_code == 200
    data = resp.get_json()["value"]
    refresh = data["refresh_token"]

    # Manually mark as revoked to simulate admin action
    from flarchitect.authentication.token_store import revoke_refresh_token

    revoke_refresh_token(refresh)

    # Attempting to use should fail
    resp2 = client_jwt_app.post("/auth/refresh", json={"refresh_token": refresh})
    assert resp2.status_code in (401, 403)



import flarchitect.core as core
import flarchitect.database as database
import flarchitect.schemas as schemas
import flarchitect.specs as specs


def test_core_exports_and_docstring():
    assert core.__doc__
    assert core.Architect.__name__ == "Architect"
    assert "Architect" in core.__all__


def test_database_exports_and_docstring():
    assert database.__doc__
    assert database.CrudService.__name__ == "CrudService"
    assert "CrudService" in database.__all__


def test_schemas_exports_and_docstring():
    assert schemas.__doc__
    assert isinstance(schemas.AutoSchema, type)
    assert "AutoSchema" in schemas.__all__


def test_specs_exports_and_docstring():
    assert specs.__doc__
    assert specs.CustomSpec.__name__ == "CustomSpec"
    assert "CustomSpec" in specs.__all__



"""Custom logging utilities providing coloured or JSON logs with verbosity control.

Adds optional structured JSON output enriched with request context when
running under Flask. Falls back to colourised human-friendly output.
"""

from __future__ import annotations

import json
import re
import time
from typing import Any

try:
    # Flask may not always be present; guard imports.
    from flask import g, has_request_context, request
except Exception:  # pragma: no cover - optional import for non-Flask usage
    def has_request_context() -> bool:  # type: ignore
        return False

    request = None  # type: ignore
    g = None  # type: ignore

from colorama import Fore, Style, init

# Initialise Colorama
init(autoreset=True)


def color_text_with_multiple_patterns(text: str) -> str:
    """Colour text wrapped in specific patterns with respective colours.

    Args:
        text: The text containing wrapped patterns.

    Returns:
        The colourised text with patterns replaced.
    """
    patterns: dict[str, tuple[str, str]] = {
        r"`(.*?)`": (Fore.YELLOW, Style.NORMAL),  # Yellow for backticks
        r"\+(.*?)\+": (Fore.RED, Style.NORMAL),  # Red for pluses
        r"--(.*?)--": (Fore.CYAN, Style.NORMAL),  # Cyan for hyphens
        r"\$(.*?)\$": (Fore.MAGENTA, Style.BRIGHT),  # Magenta for dollars
        r"\|(.*?)\|": (Fore.GREEN, Style.BRIGHT),  # Green for pipes
    }

    def replace_with_color(match: re.Match[str], color: str, style: str) -> str:
        return f"{color}{style}{match.group(1)}{Style.RESET_ALL}"

    for pattern, (color, style) in patterns.items():
        text = re.sub(
            pattern,
            lambda match, color=color, style=style: replace_with_color(match, color, style),
            text,
        )

    return text


class CustomLogger:
    """Simple logger with verbosity-based level control and JSON mode."""

    def __init__(self, verbosity_level: int = 0) -> None:
        self.verbosity_level = verbosity_level
        self.json_mode: bool = False

    def _emit_text(self, text: str) -> None:
        print(color_text_with_multiple_patterns(text))

    def _emit_json(self, payload: dict[str, Any]) -> None:
        print(json.dumps(payload, separators=(",", ":")))

    def _context(self) -> dict[str, Any]:
        if has_request_context():  # pragma: no cover - integration behaviour
            ctx: dict[str, Any] = {
                "method": request.method,
                "path": request.path,
            }
            # Best-effort request id and timing
            try:
                ctx["request_id"] = getattr(g, "request_id", None)
                start = getattr(g, "_flarch_req_start", None)
                if start is not None:
                    ctx["latency_ms"] = int((time.perf_counter() - start) * 1000)
            except Exception:
                pass
            return ctx
        return {}

    def _log_with_prefix(self, level: int, message: str, prefix: str, color: str | None = None) -> None:
        """Internal helper to log with a prefix and optional colour or JSON."""
        if level <= self.verbosity_level:
            if self.json_mode:
                payload = {
                    "event": prefix.lower(),
                    "lvl": level,
                    "message": message,
                }
                payload.update(self._context())
                self._emit_json(payload)
            else:
                prefix_text = f"{prefix} {level}: ".ljust(10)
                if color:
                    prefix_text = f"{color}{prefix_text}{Style.RESET_ALL}"
                self._emit_text(prefix_text + message)

    def log(self, level: int, message: str) -> None:
        """Log a message if its level is less than or equal to the current verbosity level."""
        self._log_with_prefix(level, message, "LOG")

    def debug(self, level: int, message: str) -> None:
        """Log a debug message if its level is less than or equal to the current verbosity level."""
        self._log_with_prefix(level, message, "DEBUG")

    def error(self, level: int, message: str) -> None:
        """Log an error message if its level is less than or equal to the current verbosity level."""
        self._log_with_prefix(level, message, "ERROR", color=Fore.RED)


logger = CustomLogger()

def get_logger() -> CustomLogger:
    """Return the module-level logger instance."""
    return logger



from __future__ import annotations

import contextlib
from collections.abc import Callable
from typing import Any

from flask import Request, Response


class PluginBase:
    """Base class for flarchitect plugins.

    Implement any of the hook methods to observe or influence behaviour.
    All hooks are optional. Default implementations are no-ops.

    Stable hook signatures (kwargs may grow but not change meaning):
    - request_started(request: Request) -> None
    - request_finished(request: Request, response: Response) -> Response | None
    - before_authenticate(context: dict[str, Any]) -> dict[str, Any] | None
    - after_authenticate(context: dict[str, Any], success: bool, user: Any | None) -> None
    - before_model_op(context: dict[str, Any]) -> dict[str, Any] | None
    - after_model_op(context: dict[str, Any], output: Any) -> Any | None
    - spec_build_started(spec: Any) -> None
    - spec_build_completed(spec_dict: dict[str, Any]) -> dict[str, Any] | None
    """

    def request_started(self, request: Request) -> None:  # pragma: no cover - default no-op
        return None

    def request_finished(self, request: Request, response: Response) -> Response | None:  # pragma: no cover - default no-op
        return None

    def before_authenticate(self, context: dict[str, Any]) -> dict[str, Any] | None:  # pragma: no cover - default no-op
        return None

    def after_authenticate(self, context: dict[str, Any], success: bool, user: Any | None) -> None:  # pragma: no cover - default no-op
        return None

    def before_model_op(self, context: dict[str, Any]) -> dict[str, Any] | None:  # pragma: no cover - default no-op
        return None

    def after_model_op(self, context: dict[str, Any], output: Any) -> Any | None:  # pragma: no cover - default no-op
        return None

    def spec_build_started(self, spec: Any) -> None:  # pragma: no cover - default no-op
        return None

    def spec_build_completed(self, spec_dict: dict[str, Any]) -> dict[str, Any] | None:  # pragma: no cover - default no-op
        return None


class PluginManager:
    """Manage registration and invocation of flarchitect plugins."""

    def __init__(self, plugins: list[PluginBase] | None = None) -> None:
        self._plugins: list[PluginBase] = plugins or []

    @staticmethod
    def _coerce(entry: Any) -> PluginBase:
        # Support instances or classes; ignore invalid entries gracefully
        if isinstance(entry, PluginBase):
            return entry
        if isinstance(entry, type) and issubclass(entry, PluginBase):
            return entry()  # type: ignore[call-arg]
        # Try callables returning a PluginBase
        if callable(entry):
            candidate = entry()
            if isinstance(candidate, PluginBase):
                return candidate
        raise TypeError("Invalid plugin entry; expected PluginBase or factory")

    @classmethod
    def from_config(cls, config_val: Any) -> PluginManager:
        plugins: list[PluginBase] = []
        if isinstance(config_val, list):
            for entry in config_val:
                try:
                    plugins.append(cls._coerce(entry))
                except Exception:
                    # Skip invalid plugins rather than breaking app startup
                    continue
        elif config_val:
            with contextlib.suppress(Exception):
                plugins.append(cls._coerce(config_val))
        return cls(plugins)

    # Dispatch helpers
    def _first_non_none(self, func: Callable[[PluginBase], Any]) -> Any:
        for p in self._plugins:
            result = func(p)
            if result is not None:
                return result
        return None

    def request_started(self, request: Request) -> None:
        for p in self._plugins:
            try:
                p.request_started(request)
            except Exception:
                continue

    def request_finished(self, request: Request, response: Response) -> Response | None:
        return self._first_non_none(lambda p: self._safe_call(p.request_finished, request, response))

    def before_authenticate(self, context: dict[str, Any]) -> dict[str, Any] | None:
        updates: dict[str, Any] | None = None
        for p in self._plugins:
            try:
                upd = p.before_authenticate(context)
                if isinstance(upd, dict):
                    context.update(upd)
                    updates = context
            except Exception:
                continue
        return updates

    def after_authenticate(self, context: dict[str, Any], success: bool, user: Any | None) -> None:
        for p in self._plugins:
            with contextlib.suppress(Exception):
                p.after_authenticate(context, success, user)

    def before_model_op(self, context: dict[str, Any]) -> dict[str, Any] | None:
        updates: dict[str, Any] | None = None
        for p in self._plugins:
            try:
                upd = p.before_model_op(context)
                if isinstance(upd, dict):
                    context.update(upd)
                    updates = context
            except Exception:
                continue
        return updates

    def after_model_op(self, context: dict[str, Any], output: Any) -> Any | None:
        out = output
        changed = False
        for p in self._plugins:
            try:
                maybe = p.after_model_op(context, out)
                if maybe is not None:
                    out = maybe
                    changed = True
            except Exception:
                continue
        return out if changed else None

    def spec_build_started(self, spec: Any) -> None:
        for p in self._plugins:
            with contextlib.suppress(Exception):
                p.spec_build_started(spec)

    def spec_build_completed(self, spec_dict: dict[str, Any]) -> dict[str, Any] | None:
        out = spec_dict
        changed = False
        for p in self._plugins:
            try:
                maybe = p.spec_build_completed(out)
                if isinstance(maybe, dict):
                    out = maybe
                    changed = True
            except Exception:
                continue
        return out if changed else None

    @staticmethod
    def _safe_call(fn: Callable, *args: Any, **kwargs: Any) -> Any:
        try:
            return fn(*args, **kwargs)
        except Exception:
            return None



"""Tests for CustomHTTPException and related helpers."""

from flask import Flask

from flarchitect.exceptions import CustomHTTPException, _handle_exception, handle_http_exception


def test_custom_http_exception_to_dict() -> None:
    exc = CustomHTTPException(400, "Invalid input")
    assert exc.to_dict() == {
        "status_code": 400,
        "status_text": "Bad Request",
        "reason": "Invalid input",
    }


def test_handle_exception_returns_response() -> None:
    app = Flask(__name__)
    with app.test_request_context():
        response = _handle_exception("Unexpected", 500, error_name="Server Error", print_exc=False)
        data = response.get_json()
        assert response.status_code == 500
        assert data["errors"] == {"error": "Unexpected", "reason": "Server Error"}
        assert data["status_code"] == 500


def test_handle_http_exception_returns_json() -> None:
    app = Flask(__name__)
    app.register_error_handler(CustomHTTPException, handle_http_exception)

    @app.get("/api/error")
    def error_route() -> dict[str, str]:  # pragma: no cover - simple route for test
        raise CustomHTTPException(401, "Auth required")

    with app.test_client() as client:
        response = client.get("/api/error")
        data = response.get_json()
        assert response.status_code == 401
        assert data["errors"] == {"error": "Unauthorized", "reason": "Auth required"}



"""Demo application showcasing flarchitect field validators.

This package provides a minimal Flask application demonstrating how
:func:`flarchitect.schemas.validators.validate_by_type` integrates with
SQLAlchemy models.  It exposes :func:`create_app` which is used by the
unit tests and can be executed directly.
"""

from .app import User, create_app, db

__all__ = ["create_app", "db", "User"]



"""Extension instances for the scaffolding application."""

from __future__ import annotations

import datetime
from typing import Any

from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import Boolean, DateTime
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from flarchitect import Architect


class BaseModel(DeclarativeBase):
    """Base model with timestamp and soft delete support."""

    created: Mapped[datetime.datetime] = mapped_column(DateTime, default=datetime.datetime.utcnow)
    updated: Mapped[datetime.datetime] = mapped_column(DateTime, default=datetime.datetime.utcnow, onupdate=datetime.datetime.utcnow)
    deleted: Mapped[bool] = mapped_column(Boolean, default=False, nullable=False)

    def get_session(*_args: Any, **_kwargs: Any):
        """Return the current database session."""

        return db.session


# Instantiate extensions globally to be initialised in the app factory.
db = SQLAlchemy(model_class=BaseModel)
schema = Architect()

__all__ = ["db", "schema", "BaseModel"]



from demo.callbacks.model.callbacks import (
    dump_callback,
    error_callback,
    final_callback,
    return_callback,
    setup_callback,
)
from demo.callbacks.model.extensions import db


class Config:
    SQLALCHEMY_DATABASE_URI = "sqlite:///:memory:"
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    API_BASE_MODEL = db.Model
    API_TITLE = "Book Shop API"
    API_VERSION = "0.1.0"
    API_VERBOSITY_LEVEL = 4
    API_DOCS_FONT = "montserrat"
    API_RATE_LIMIT = "5 per minute"
    SECRET_KEY = "8Kobns1_vnmg3rxnr0RZpkF4D1s"
    API_SETUP_CALLBACK = setup_callback
    API_RETURN_CALLBACK = return_callback
    API_FINAL_CALLBACK = final_callback
    API_ERROR_CALLBACK = error_callback
    API_DUMP_CALLBACK = dump_callback



from __future__ import annotations

import importlib
import json
import threading
import time
from collections import defaultdict
from dataclasses import dataclass
from queue import Empty, Queue
from typing import Any

from flarchitect.logging import logger


@dataclass
class _Subscription:
    topic: str
    queue: Queue


class _EventBus:
    """Very small in-memory pub/sub used for WebSocket broadcasting.

    - Topics are free-form strings (e.g. "all", "author", "book").
    - Subscribers receive JSON-serialisable dict messages.
    - This is process-local and non-durable; intended for development and
      single-process deployments. For production, prefer a real broker
      (Redis, NATS, etc.) and swap out publish/subscribe implementations.
    """

    def __init__(self) -> None:
        self._subs: defaultdict[str, set[Queue]] = defaultdict(set)
        self._lock = threading.Lock()

    def subscribe(self, topic: str) -> _Subscription:
        q: Queue = Queue()
        with self._lock:
            self._subs[topic].add(q)
        logger.debug(5, f"Subscribed queue to topic '{topic}'")
        return _Subscription(topic=topic, queue=q)

    def unsubscribe(self, sub: _Subscription) -> None:
        with self._lock:
            queues = self._subs.get(sub.topic)
            if queues and sub.queue in queues:
                queues.remove(sub.queue)
                if not queues:
                    self._subs.pop(sub.topic, None)
        logger.debug(5, f"Unsubscribed queue from topic '{sub.topic}'")

    def publish(self, topic: str, message: dict[str, Any]) -> None:
        with self._lock:
            # deliver to explicit topic and to 'all'
            targets = set(self._subs.get(topic, set())) | set(self._subs.get("all", set()))
        import contextlib
        for q in targets:
            with contextlib.suppress(Exception):  # pragma: no cover - defensive
                q.put_nowait(message)


_BUS = _EventBus()


def broadcast_change(*, model: Any | None, method: str, payload: Any, id: Any | None = None, many: bool = False) -> None:
    """Publish a change event to WebSocket subscribers.

    Args:
        model: SQLAlchemy model class the change applies to.
        method: The HTTP method that triggered the change (GET/POST/PATCH/DELETE).
        payload: Marshmallow-serialised output returned to the client.
        id: Optional primary key for single-object operations.
        many: Whether the payload contains multiple items.
    """
    try:
        model_name = model.__name__.lower() if model is not None else "unknown"
        message = {
            "ts": int(time.time() * 1000),
            "model": model_name,
            "method": method.upper(),
            "id": id,
            "many": bool(many),
            "payload": payload,
        }
        _BUS.publish(model_name, message)
        _BUS.publish("all", message)
        logger.debug(5, f"Broadcasted WS message for '{model_name}' {method}")
    except Exception as exc:  # pragma: no cover - best effort only
        logger.debug(4, f"WebSocket broadcast skipped: {exc}")


def init_websockets(architect: Any, path: str = "/ws") -> None:
    """Attach a minimal WebSocket endpoint if Flask-Sock is available.

    The endpoint accepts optional query parameters:
      - topic: subscribe to a specific model topic (defaults to 'all')

    Message format is JSON lines, one per event as emitted by ``broadcast_change``.

    This function is a no-op if the optional dependency ``flask_sock`` is not
    installed.
    """
    try:
        spec = importlib.util.find_spec("flask_sock")
        if spec is None:
            logger.debug(3, "flask_sock not installed; WebSocket endpoint not registered")
            return

        from flask import request
        from flask_sock import Sock  # type: ignore

        sock = Sock(architect.app)

        @sock.route(path)
        def ws(sock):  # type: ignore
            topic = (request.args.get("topic") or "all").lower()
            sub = _BUS.subscribe(topic)
            try:
                while True:
                    try:
                        msg = sub.queue.get(timeout=1.0)
                    except Empty:
                        # keep connection alive; allow client pings to be handled
                        continue
                    try:
                        sock.send(json.dumps(msg))
                    except Exception:
                        break
            finally:
                _BUS.unsubscribe(sub)

        logger.log(2, f"Registered WebSocket route at '{path}' using flask_sock")
    except Exception as exc:  # pragma: no cover - defensive
        logger.debug(3, f"Failed to initialise WebSocket endpoint: {exc}")



"""JWT error handling tests."""

from __future__ import annotations

import datetime
from collections.abc import Generator
from types import SimpleNamespace
from unittest.mock import Mock

import jwt
import pytest
from flask import Flask
from pytest import MonkeyPatch

from flarchitect.authentication.jwt import decode_token, refresh_access_token
from flarchitect.exceptions import CustomHTTPException


@pytest.fixture()
def app_ctx() -> Generator[Flask, None, None]:
    """Provide a Flask application context for JWT operations."""
    app = Flask(__name__)
    with app.app_context():
        yield app


@pytest.fixture()
def secret_key() -> str:
    """Return the secret key used for encoding test JWTs."""
    return "secret"


@pytest.fixture()
def refresh_secret() -> str:
    """Return the secret key for refresh token encoding."""
    return "refresh"


@pytest.fixture()
def expired_token(secret_key: str) -> str:
    """Create an expired JWT for testing."""
    payload = {
        "sub": "user",
        "iat": datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(minutes=2),
        "exp": datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(minutes=1),
    }
    return jwt.encode(payload, secret_key, algorithm="HS256")


@pytest.fixture()
def malformed_token() -> str:
    """Return a malformed JWT string."""
    return "invalid.token"


@pytest.fixture()
def valid_refresh_token(refresh_secret: str) -> str:
    """Generate a valid refresh token for testing."""
    payload = {
        "sub": "user",
        "iat": datetime.datetime.now(datetime.timezone.utc),
        "exp": datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(minutes=30),
    }
    return jwt.encode(payload, refresh_secret, algorithm="HS256")


def test_decode_token_expired(app_ctx: Flask, expired_token: str, secret_key: str) -> None:
    """Ensure ``decode_token`` raises ``CustomHTTPException`` for expired tokens."""
    with pytest.raises(CustomHTTPException) as exc_info:
        decode_token(expired_token, secret_key)
    assert exc_info.value.status_code == 401
    assert exc_info.value.reason == "Token has expired"


def test_decode_token_malformed(app_ctx: Flask, malformed_token: str, secret_key: str) -> None:
    """Ensure ``decode_token`` raises ``CustomHTTPException`` for malformed tokens."""
    with pytest.raises(CustomHTTPException) as exc_info:
        decode_token(malformed_token, secret_key)
    assert exc_info.value.status_code == 401
    assert exc_info.value.reason == "Invalid token"


def test_refresh_access_token_deletes_expired_token(
    app_ctx: Flask,
    valid_refresh_token: str,
    refresh_secret: str,
    monkeypatch: MonkeyPatch,
) -> None:
    """Verify expired refresh tokens are deleted and return ``403``."""
    app_ctx.config["REFRESH_SECRET_KEY"] = refresh_secret

    past_time = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(minutes=1)
    stored = SimpleNamespace(expires_at=past_time, user_lookup="user", user_pk="1")

    delete_mock = Mock()
    monkeypatch.setattr("flarchitect.authentication.jwt.delete_refresh_token", delete_mock)
    monkeypatch.setattr("flarchitect.authentication.jwt.get_refresh_token", lambda token: stored)

    with pytest.raises(CustomHTTPException) as exc_info:
        refresh_access_token(valid_refresh_token)

    assert exc_info.value.status_code == 403
    assert exc_info.value.reason == "Invalid or expired refresh token"
    delete_mock.assert_called_once_with(valid_refresh_token)


def test_refresh_access_token_missing_secret_key(app_ctx: Flask, valid_refresh_token: str, monkeypatch: MonkeyPatch) -> None:
    """Ensure a ``500`` is raised when ``REFRESH_SECRET_KEY`` is missing."""
    monkeypatch.delenv("REFRESH_SECRET_KEY", raising=False)
    app_ctx.config.pop("REFRESH_SECRET_KEY", None)

    get_mock = Mock()
    delete_mock = Mock()
    monkeypatch.setattr("flarchitect.authentication.jwt.get_refresh_token", get_mock)
    monkeypatch.setattr("flarchitect.authentication.jwt.delete_refresh_token", delete_mock)

    with pytest.raises(CustomHTTPException) as exc_info:
        refresh_access_token(valid_refresh_token)

    assert exc_info.value.status_code == 500
    assert exc_info.value.reason == "REFRESH_SECRET_KEY missing"
    get_mock.assert_not_called()
    delete_mock.assert_not_called()



"""Example callback functions for the callbacks demo application."""

from __future__ import annotations

from typing import Any


def setup_callback(*, model, **kwargs) -> dict[str, Any]:
    """Example setup callback used in the demo.

    Args:
        model: The model class being accessed.
        **kwargs: Additional keyword arguments passed from the framework.

    Returns:
        Dict[str, Any]: Potentially modified keyword arguments.
    """
    return kwargs


def return_callback(*, model, output, **kwargs) -> dict[str, Any]:
    """Example return callback used in the demo.

    Args:
        model: The model class being accessed.
        output: The value returned from the database operation.
        **kwargs: Additional keyword arguments.

    Returns:
        Dict[str, Any]: A dictionary containing the ``output`` key.
    """
    return {"output": output}


def final_callback(data: dict[str, Any]) -> dict[str, Any]:
    """Attach a marker before the response is sent."""
    data["finalized"] = True
    return data


def error_callback(error: str, status_code: int, value: Any) -> None:
    """Handle errors raised during a request."""
    _ = (error, status_code, value)  # Placeholder for real logging.


def dump_callback(data: dict[str, Any], **kwargs) -> dict[str, Any]:
    """Modify serialised data prior to returning it to the client."""
    data["demo"] = True
    return data



import json

import pytest

from demo.model_extension.model import create_app as create_app_models
from flarchitect.utils.core_utils import convert_case


@pytest.mark.parametrize("case", ["camel", "pascal"])
def test_nested_schema_refs_follow_case(case: str) -> None:
    app = create_app_models({"API_SCHEMA_CASE": case})
    client = app.test_client()
    spec = client.get("/apispec.json").json
    spec_str = json.dumps(spec)
    expected = convert_case("Category", case)
    unexpected = "Category" if expected == "category" else "category"
    assert f"#/components/schemas/{expected}" in spec_str
    assert f"#/components/schemas/{unexpected}" not in spec_str



"""Simple in-memory cache backend for environments without :mod:`flask_caching`.

This fallback cache implements a very small subset of the ``flask_caching``
interface required by the tests.  It provides an in-memory key/value store with
per-item timeouts and a ``cached`` decorator used by :class:`~flarchitect.core.routes.RouteCreator`.
"""

from __future__ import annotations

import time
from collections.abc import Callable
from functools import wraps
from typing import Any

from flask import request


class SimpleCache:
    """In-memory cache with an API similar to ``flask_caching``'s ``Cache``.

    Args:
        default_timeout: Default cache timeout in seconds.

    Methods:
        clear: Remove all entries from the cache.
    """

    def __init__(self, default_timeout: int = 300) -> None:
        self.default_timeout = default_timeout
        self._cache: dict[str, tuple[float, Any]] = {}

    def clear(self) -> None:
        """Clear all data from the cache."""

        self._cache.clear()

    def init_app(self, app) -> None:  # type: ignore[no-untyped-def]
        """Initialise the cache for a Flask application.

        This implementation stores data in process memory, so no explicit
        initialisation work is required.  The method is provided to mirror the
        ``flask_caching`` API.
        """

    def _expires(self, timeout: int | None) -> float:
        """Return an absolute expiry timestamp.

        Args:
            timeout: Timeout in seconds. ``None`` uses ``self.default_timeout``.

        Returns:
            The epoch time when the cache entry should expire.
        """

        return time.time() + (timeout if timeout is not None else self.default_timeout)

    def _purge(self) -> None:
        """Remove expired entries from the cache."""

        now = time.time()
        expired = [key for key, (expires, _value) in self._cache.items() if expires < now]
        for key in expired:
            self._cache.pop(key, None)

    def get(self, key: str) -> Any | None:
        """Retrieve a cached value if present and not expired.

        Args:
            key: Cache key to retrieve.

        Returns:
            The cached value if found and valid; otherwise ``None``.
        """

        self._purge()
        record = self._cache.get(key)
        if record is None:
            return None
        _expires, value = record
        return value

    def set(self, key: str, value: Any, timeout: int | None = None) -> None:
        """Store ``value`` in the cache with the given ``timeout``."""

        self._cache[key] = (self._expires(timeout), value)

    def cached(self, timeout: int | None = None) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        """Decorator to cache view function responses.

        Args:
            timeout: Time in seconds before the cache entry expires.

        Returns:
            A decorator that caches the wrapped function's response keyed by the
            request's full path.
        """

        def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
            @wraps(func)
            def wrapper(*args: Any, **kwargs: Any) -> Any:
                key = request.full_path  # type: ignore[union-attr]
                cached = self.get(key)
                if cached is not None:
                    return cached
                value = func(*args, **kwargs)
                self.set(key, value, timeout)
                return value

            return wrapper

        return decorator



import pytest
from flask.testing import FlaskClient

from demo.basic_factory.basic_factory import create_app


@pytest.fixture
def client() -> FlaskClient:
    app = create_app({})
    return app.test_client()


def test_default_includes_fields(client):
    """By default ``response_ms`` and ``total_count`` are present."""
    resp = client.get("/api/books/1").json
    assert "response_ms" in resp
    assert "total_count" in resp


def test_disable_response_ms():
    app = create_app({"API_DUMP_RESPONSE_MS": False})
    client = app.test_client()
    resp = client.get("/api/books/1").json
    assert "response_ms" not in resp
    assert "total_count" in resp


def test_disable_total_count():
    app = create_app({"API_DUMP_TOTAL_COUNT": False})
    client = app.test_client()
    resp = client.get("/api/books/1").json
    assert "total_count" not in resp
    assert "response_ms" in resp


def test_request_id_not_in_body_by_default():
    """By default, request_id is only returned via header, not body."""
    app = create_app({})
    client = app.test_client()
    r = client.get("/api/books/1")
    body = r.get_json()
    assert "request_id" not in body
    # But header should be present
    assert r.headers.get("X-Request-ID")


def test_request_id_in_body_when_enabled():
    """request_id appears in body when API_DUMP_REQUEST_ID=True."""
    app = create_app({"API_DUMP_REQUEST_ID": True})
    client = app.test_client()
    r = client.get("/api/books/1")
    body = r.get_json()
    rid_header = r.headers.get("X-Request-ID")
    assert "request_id" in body
    assert isinstance(body["request_id"], str)
    # Value should match header correlation id
    assert body["request_id"] == rid_header



"""Demo application showcasing response caching.

The app exposes a single ``Author`` model with a generated REST API. GET
responses are cached using the backend configured via ``API_CACHE_TYPE``.
``API_CACHE_TIMEOUT`` controls how long each cache entry lives. When the
``flask_caching`` package is installed the demo uses its ``Cache`` class; if
not, it falls back to :class:`flarchitect.core.simple_cache.SimpleCache`.
"""

from __future__ import annotations

import time
from typing import Any

from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import Integer, String
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from flarchitect import Architect


class BaseModel(DeclarativeBase):
    """Base model providing ``get_session`` for flarchitect."""

    def get_session(*args: Any) -> Any:  # type: ignore[no-untyped-def]
        return db.session


db = SQLAlchemy(model_class=BaseModel)


class Author(db.Model):
    """Simple author model used for demonstration."""

    __tablename__ = "author"

    class Meta:
        tag_group = "People/Companies"
        tag = "Author"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    first_name: Mapped[str] = mapped_column(String)


def create_app(config: dict | None = None) -> Flask:
    """Application factory enabling response caching.

    The returned app caches GET responses. ``API_CACHE_TYPE`` selects the cache
    backend while ``API_CACHE_TIMEOUT`` defines the expiration in seconds.

    Args:
        config: Optional mapping to override default configuration.

    Returns:
        Configured :class:`~flask.Flask` instance.
    """

    app = Flask(__name__)
    app.config.update(
        SQLALCHEMY_DATABASE_URI="sqlite:///:memory:",
        API_TITLE="Caching Demo",
        API_VERSION="1.0",
        API_BASE_MODEL=db.Model,
        API_CACHE_TYPE="SimpleCache",
        API_CACHE_TIMEOUT=1,
    )
    if config:
        app.config.update(config)

    db.init_app(app)
    with app.app_context():
        db.create_all()
        db.session.add(Author(first_name="Initial"))
        db.session.commit()

        architect = Architect(app)

        @app.route("/time")
        @architect.cache.cached()
        def current_time() -> dict[str, float]:
            """Return the current epoch time.

            The value is cached, so repeated calls within the timeout return
            the same response. After ``API_CACHE_TIMEOUT`` seconds, the cache is
            invalidated and a new timestamp is generated.
            """

            return {"time": time.time()}

    return app


if __name__ == "__main__":  # pragma: no cover - manual execution only
    create_app().run(debug=True)



from __future__ import annotations

import os
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin
from apispec.ext.marshmallow.common import make_schema_key, resolve_schema_instance
from flask import Blueprint, Flask, Response, current_app, redirect, request, session
from flask_login import current_user, login_user
from marshmallow import Schema
from sqlalchemy.orm import DeclarativeBase

from flarchitect.core.routes import (
    create_params_from_rule,
    create_query_params_from_rule,
    find_rule_by_function,
)
from flarchitect.logging import logger
from flarchitect.specs.utils import (
    append_parameters,
    convert_path_to_openapi,
    handle_authorisation,
    initialise_spec_template,
    schema_name_resolver,
    scrape_extra_info_from_spec_data,
)
from flarchitect.utils.config_helpers import get_config_or_model_meta
from flarchitect.utils.general import (
    AttributeInitialiserMixin,
    generate_readme_html,
    get_html_path,
    make_base_dict,
    manual_render_absolute_template,
    pretty_print_dict,
    search_all_keys,
)
from flarchitect.utils.response_helpers import create_response

if TYPE_CHECKING:  # pragma: no cover - imported only for type hints
    from flarchitect import Architect


class CustomSpec(APISpec, AttributeInitialiserMixin):
    """A subclass of APISpec for specifying tag groups with extended features."""

    app: Flask
    architect: Architect  # The architect object

    spec_groups: dict[str, list[dict[str, str | list[str]]]] | None = None
    api_title: str | None = ""
    api_version: str | None = ""
    api_description: str | None = None
    api_logo_url: str | None = None
    api_logo_background: str | None = None
    api_keywords: list[str] | None = None
    create_docs: bool | None = True
    documentation_url_prefix: str | None = None
    documentation_url: str | None = "/docs"

    def __init__(self, app: Flask, architect: Architect, *args, **kwargs):
        """Initialises the CustomSpec object.

        Args:
            app (Flask): The Flask application.
            architect (Architect): The Architect object.
            *args: Positional arguments passed to the parent class.
            **kwargs: Keyword arguments passed to the parent class.
        """
        self.app = app
        self.architect = architect
        # initialise per-instance containers to avoid shared mutable defaults
        self.spec_groups: dict[str, list[dict[str, str | list[str]]]] = {"x-tagGroups": []}
        self.api_keywords: list[str] = []
        super().__init__(*args, **self._prepare_api_spec_data(**kwargs))

        # Reset MarshmallowPlugin schema refs to avoid cross-app/test bleed.
        import contextlib
        with contextlib.suppress(Exception):
            plugin = next((p for p in self.plugins if isinstance(p, MarshmallowPlugin)), None)
            if plugin and getattr(getattr(plugin, "converter", None), "refs", None) is not None:
                plugin.converter.refs.clear()  # type: ignore[attr-defined]

        if self._should_create_docs():
            self.architect.api_spec = self
            self._create_specification_blueprint()
            register_routes_with_spec(self.architect, self.architect.route_spec)

    def to_dict(self) -> dict:
        """Converts the API specification to a dictionary.

        Returns:
            dict: The API specification as a dictionary.
        """
        # Notify plugins that spec build is starting/completing
        import contextlib
        with contextlib.suppress(Exception):
            self.architect.plugins.spec_build_started(self)
        spec_dict = super().to_dict()
        with contextlib.suppress(Exception):
            maybe = self.architect.plugins.spec_build_completed(spec_dict)
            if isinstance(maybe, dict):
                spec_dict = maybe
        if self.spec_groups:
            spec_dict.update(self.spec_groups)
        return spec_dict

    def _prepare_api_spec_data(self, **kwargs) -> dict:
        """Prepares the data required to initialise the API spec.

        Returns:
            dict: Data for initialising the API spec.
        """
        api_description = self._get_api_description()
        api_spec_data = {
            "openapi_version": "3.0.2",
            "plugins": [MarshmallowPlugin(schema_name_resolver=schema_name_resolver)],
            "title": self._get_config("API_TITLE", "My API"),
            "version": self._get_config("API_VERSION", "1.0.0"),
            "info": {
                "description": api_description,
                **self._get_contact_info(),
                **self._get_license_info(),
                **self._get_logo_info(),
            },
            "components": {
                "securitySchemes": {
                    "bearerAuth": {
                        "type": "http",
                        "scheme": "bearer",
                        "bearerFormat": "JWT",
                    }
                }
            },
            **self._get_servers_info(),
        }
        return {**api_spec_data, **kwargs}

    def _get_api_description(self) -> str:
        """Generates API description in HTML format.

        Returns:
            str: The API description in HTML format.
        """
        desc_path = get_config_or_model_meta(
            "API_DESCRIPTION",
            default=os.path.join(self.architect.get_templates_path(), "base_readme.MD"),
        )
        if os.path.isfile(desc_path):
            base_model = get_config_or_model_meta("API_BASE_MODEL")
            return generate_readme_html(
                desc_path,
                config=self.architect.app.config,
                api_output_example=pretty_print_dict(make_base_dict()),
                has_rate_limiting=search_all_keys(base_model, "API_RATE_LIMIT"),
                has_auth=search_all_keys(base_model, "API_AUTHENTICATE_METHOD"),
            )
        return desc_path

    def _get_config(self, key: str, default: str | None = None) -> str | None:
        """Fetches configuration or model metadata.

        Args:
            key (str): The configuration key to retrieve.
            default (Optional[str]): Default value if key not found.

        Returns:
            Optional[str]: The value for the given configuration key.
        """
        return get_config_or_model_meta(key, default=default)

    def _get_contact_info(self) -> dict[str, dict[str, str | None]]:
        """Retrieves contact information for the API spec.

        Returns:
            dict: A dictionary containing the contact information.
        """
        contact_info = {k: self._get_config(f"API_CONTACT_{k.upper()}") for k in ["name", "email", "url"]}
        return {"contact": {k: v for k, v in contact_info.items() if v}} if any(contact_info.values()) else {}

    def _get_license_info(self) -> dict[str, dict[str, str | None]]:
        """Retrieves license information for the API spec.

        Returns:
            dict: A dictionary containing the license information.
        """
        license_info = {k: self._get_config(f"API_LICENCE_{k.upper()}") for k in ["name", "url"]}
        return {"license": {k: v for k, v in license_info.items() if v}} if any(license_info.values()) else {}

    def _get_servers_info(self) -> dict[str, list[str] | None]:
        """Retrieves server URLs for the API spec.

        Returns:
            dict: A dictionary containing server information.
        """
        servers = self._get_config("API_SERVER_URLS")
        return {"servers": servers} if servers else {}

    def _get_logo_info(self) -> dict[str, dict[str, str | None]]:
        """Retrieves logo information for the API spec.

        Returns:
            dict: A dictionary containing the logo information.
        """
        logo_url = self._get_config("API_LOGO_URL")
        if logo_url:
            return {
                "x-logo": {
                    "url": logo_url,
                    "backgroundColor": self._get_config("API_LOGO_BACKGROUND", "#ffffff"),
                    "altText": f"{self._get_config('API_TITLE', 'My API')} logo.",
                }
            }
        return {}

    def _should_create_docs(self) -> bool:
        """Determines whether to create API documentation.

        Returns:
            bool: True if documentation should be created, False otherwise.
        """
        return get_config_or_model_meta("API_CREATE_DOCS", default=True)

    def validate_init_apispec_args(self) -> None:
        """Validates the initialisation arguments for the API spec.

        Raises:
            ValueError: If any required argument is invalid.
        """
        if not isinstance(self.api_title, str) or not self.api_title:
            raise ValueError("API title must be a non-empty string.")
        if not isinstance(self.api_version, str) or not self.api_version:
            raise ValueError("API version must be a non-empty string.")

    def set_xtags_group(self, tag_name: str, group_name: str) -> None:
        """Adds a tag to a tag-group, creating the group if it doesn't exist.

        Args:
            tag_name (str): The tag to add.
            group_name (str): The group to add the tag to.

        Example:
            >>> spec.set_xtags_group("Users", "Authentication")
        """
        if self.spec_groups is None:
            self.spec_groups = {"x-tagGroups": []}

        tag_groups = self.spec_groups.setdefault("x-tagGroups", [])

        for group in tag_groups:
            if group["name"] == group_name:
                if tag_name not in group["tags"]:
                    group["tags"].append(tag_name)
                return
        tag_groups.append({"name": group_name, "tags": [tag_name]})

    def _create_specification_blueprint(self) -> None:
        """Sets up the blueprint to serve the API specification and documentation.

        The HTML template directory is resolved dynamically using
        :func:`~flarchitect.utils.general.get_html_path` to avoid reliance on
        global state.
        """

        html_path = get_html_path()

        specification = Blueprint(
            "specification",
            __name__,
            static_folder=html_path,
            url_prefix=self.documentation_url_prefix or self._get_config("DOCUMENTATION_URL_PREFIX", "/"),
        )

        documentation_url = get_config_or_model_meta("API_DOCUMENTATION_URL", default="/docs")
        docs_password = get_config_or_model_meta("API_DOCUMENTATION_PASSWORD", default=None)
        docs_require_auth = get_config_or_model_meta("API_DOCUMENTATION_REQUIRE_AUTH", default=False)
        auth_method = get_config_or_model_meta("API_AUTHENTICATE_METHOD", default=None)
        user_model = get_config_or_model_meta("API_USER_MODEL", default=None)

        def _ensure_docs_access(
            json_only: bool = False,
        ) -> Response | tuple[str, int] | None:
            """Validate optional documentation authentication.

            Args:
                json_only (bool): When ``True`` return a JSON error response instead of HTML.

            Returns:
                Optional[Union[Response, tuple[str, int]]]: ``None`` if access is granted, otherwise a response prompting for login.
            """

            if not (docs_password or docs_require_auth):
                return None

            if session.get("docs_authenticated") or getattr(current_user, "is_authenticated", False):
                return None

            if request.method == "POST" and not json_only:
                username = request.form.get("username")
                password = request.form.get("password", "")

                if docs_password and password == docs_password:
                    session["docs_authenticated"] = True
                    return redirect(request.path)

                if auth_method and user_model and username and password:
                    lookup_field = get_config_or_model_meta("API_USER_LOOKUP_FIELD", model=user_model, default=None)
                    check_method = get_config_or_model_meta("API_CREDENTIAL_CHECK_METHOD", model=user_model, default=None)
                    usr = user_model.query.filter(getattr(user_model, lookup_field) == username).first()
                    if usr and getattr(usr, check_method)(password):
                        session["docs_authenticated"] = True
                        login_user(usr)
                        return redirect(request.path)

                error = "Invalid credentials"
            else:
                error = None

            if json_only:
                return create_response(status=401, errors="Unauthorized")

            docs_style = get_config_or_model_meta("API_DOCS_STYLE", default="redoc").lower()
            allow_username = bool(auth_method and user_model)
            html = manual_render_absolute_template(
                os.path.join(self.architect.get_templates_path(), "docs_login.html"),
                config=self.app.config,
                docs_style=docs_style,
                allow_username=allow_username,
                error=error,
            )
            return html, 200

        @specification.route("apispec.json")
        def get_swagger_spec() -> dict | Response | tuple[str, int]:
            """Serve the Swagger spec as JSON.

            Returns:
                Union[dict, Response, tuple[str, int]]: The Swagger spec or an authentication response.
            """

            unauthorized = _ensure_docs_access(json_only=True)
            if unauthorized:
                return unauthorized
            return self.architect.to_api_spec()

        @specification.route("swagger.json")
        def get_legacy_swagger_spec() -> dict | Response | tuple[str, int]:
            """Serve the Swagger spec at legacy ``swagger.json`` endpoint."""

            return get_swagger_spec()

        get_swagger_spec._auth_disabled = True

        @specification.route(documentation_url, methods=["GET", "POST"])
        @specification.route(documentation_url + "/", methods=["GET", "POST"])
        def get_docs() -> str | Response | tuple[str, int]:
            """Serve the API documentation page.

            Returns:
                Union[str, Response, tuple[str, int]]: HTML documentation or an authentication response.
            """

            unauthorized = _ensure_docs_access()
            if unauthorized:
                return unauthorized

            custom_headers = get_config_or_model_meta("API_DOCUMENTATION_HEADERS", default="") or self._get_config("API_DOC_HTML_HEADERS", "")
            docs_style = get_config_or_model_meta("API_DOCS_STYLE", default="redoc").lower()

            template_name = "swagger.html" if docs_style == "swagger" else "apispec.html"
            return manual_render_absolute_template(
                os.path.join(self.architect.get_templates_path(), template_name),
                config=self.app.config,
                custom_headers=custom_headers,
            )

        get_docs._auth_disabled = True

        self.architect.app.register_blueprint(specification)

        prefix = self.documentation_url_prefix or self._get_config("DOCUMENTATION_URL_PREFIX", "/")
        docs_url = documentation_url if documentation_url.startswith("/") else f"/{documentation_url}"
        full_url = f"{prefix.rstrip('/')}{docs_url}"
        server_name = self.app.config.get("SERVER_NAME")
        if server_name:
            full_url = f"http://{server_name}{full_url}"
        logger.log(1, f"API documentation available at |{full_url}|")


def generate_swagger_spec(
    http_method: str,
    f: Callable,
    input_schema: Schema | None = None,
    output_schema: Schema | None = None,
    model: DeclarativeBase | None = None,
    query_params: list[dict[str, Any]] | None = None,
    path_params: list[dict[str, Any]] | None = None,
    many: bool = False,
    error_responses: list[int] | None = None,
) -> dict[str, Any]:
    """Generates a Swagger spec for an endpoint.

    Args:
        http_method (str): The HTTP method.
        f (Callable): The function associated with the endpoint.
        input_schema (Optional[Schema]): The schema for the input data.
        output_schema (Optional[Schema]): The schema for the output data.
        model (Optional[DeclarativeBase]): The SQLAlchemy model.
        query_params (Optional[List[Dict[str, Any]]]): Query parameters.
        path_params (Optional[List[Dict[str, Any]]]): Path parameters.
        many (bool): Whether the endpoint returns multiple items.
        error_responses (Optional[List[int]]): List of error response status codes.

    Returns:
        Dict[str, Any]: The generated Swagger spec.
    """
    spec = current_app.extensions["flarchitect"].api_spec

    register_schemas(spec, input_schema, output_schema)
    rate_limit = get_config_or_model_meta(
        "API_RATE_LIMIT",
        model=model,
        input_schema=input_schema,
        output_schema=output_schema,
        default=False,
    )
    spec_template = initialise_spec_template(http_method, many, rate_limit, error_responses)

    append_parameters(
        spec_template,
        query_params or [],
        path_params or [],
        http_method,
        input_schema,
        output_schema,
        model,
        many,
    )
    handle_authorisation(f, spec_template)

    return spec_template


def register_schemas(
    spec: APISpec,
    input_schema: Schema,
    output_schema: Schema | None = None,
    force_update: bool = False,
):
    """Registers schemas with the apispec object.

    Args:
        spec (APISpec): APISpec object to register schemas with.
        input_schema (Schema): Input schema to register.
        output_schema (Schema): Output schema to register.
        force_update (bool): If True, will update the schema even if it already exists.

    Returns:
        None
    """

    # ``MarshmallowPlugin`` keeps track of registered schemas. Inspect the
    # plugin to avoid adding the same schema class twice, which would trigger
    # warnings from ``apispec``.
    plugin = next((p for p in spec.plugins if isinstance(p, MarshmallowPlugin)), None)
    registered_refs = getattr(getattr(plugin, "converter", None), "refs", {}) if plugin else {}

    for schema in [input_schema, output_schema]:
        if schema:
            schema_instance = resolve_schema_instance(schema)
            schema_name = schema_name_resolver(schema_instance)

            schema_key = make_schema_key(schema_instance)
            existing_ref = registered_refs.get(schema_key)
            if existing_ref and not force_update:
                if existing_ref != schema_name:
                    spec.components.schemas[schema_name] = spec.components.schemas.pop(existing_ref)
                    registered_refs[schema_key] = schema_name
                continue

            existing_schema = spec.components.schemas.get(schema_name)
            if existing_schema and force_update:
                spec.components.schemas[schema_name] = schema_instance
            elif not existing_schema:
                spec.components.schema(schema_name, schema=schema_instance)

            registered_refs[schema_key] = schema_name


def register_routes_with_spec(architect: Architect, route_spec: list[dict[str, Any]] | None = None) -> None:
    """Register routes and schemas with the API spec.

    Args:
        architect: The :class:`~flarchitect.Architect` instance.
        route_spec: Optional list of routes and schemas to register with the
            apispec. Defaults to no action when ``None`` or empty.

    Returns:
        None
    """

    if not route_spec:
        return

    for route_info in route_spec:
        with architect.app.test_request_context():
            f = route_info["function"]
            rule = find_rule_by_function(architect, f)

            if rule:
                methods = rule.methods - {"OPTIONS", "HEAD"}
                for http_method in methods:
                    summary = route_info.get("summary")
                    route_info = scrape_extra_info_from_spec_data(route_info, method=http_method, summary=summary)
                    path = rule.rule

                    output_schema = route_info.get("output_schema")
                    input_schema = route_info.get("input_schema")
                    model = route_info.get("model")
                    description = route_info.get("description")
                    summary = route_info.get("summary")
                    custom_query_params = route_info.get("query_params")
                    tag = route_info.get("tag")
                    many = route_info.get("multiple")
                    error_responses = route_info.get("error_responses")

                    path_params = create_params_from_rule(model, rule, output_schema)
                    final_query_params = create_query_params_from_rule(rule, methods, output_schema, many, model, custom_query_params)

                    endpoint_spec = generate_swagger_spec(
                        http_method,
                        f,
                        input_schema=input_schema,
                        output_schema=output_schema,
                        model=model,
                        query_params=final_query_params,
                        path_params=path_params,
                        many=many,
                        error_responses=error_responses,
                    )

                    endpoint_spec["tags"] = [tag]
                    if not any(t.get("name") == tag for t in architect.api_spec._tags):
                        architect.api_spec.tag({"name": tag})

                    if route_info.get("group_tag"):
                        architect.api_spec.set_xtags_group(tag, route_info["group_tag"])

                    if summary:
                        endpoint_spec["summary"] = summary
                    if description:
                        endpoint_spec["description"] = description

                    architect.api_spec.path(
                        path=convert_path_to_openapi(path),
                        operations={http_method.lower(): endpoint_spec},
                    )



"""Role-based access control decorators."""

from collections.abc import Callable
from functools import wraps
from typing import Any, TypeVar, cast

from flarchitect.authentication.user import get_current_user
from flarchitect.exceptions import CustomHTTPException

F = TypeVar("F", bound=Callable[..., Any])


def require_roles(*roles: str, any_of: bool = False) -> Callable[[F], F]:
    """Enforce role-based access on the decorated function.

    Why/How:
        Validates that the authenticated user exposes required roles on
        ``current_user.roles``. When ``any_of`` is True, possession of any
        listed role is sufficient; otherwise all roles are required.

    Args:
        *roles: Role names to check against ``current_user.roles``.
        any_of: When True, allow access if any role matches; when False, all
            roles must be present.

    Returns:
        A decorator enforcing the role check.

    Raises:
        CustomHTTPException: 401 when unauthenticated, 403 when roles do not
            satisfy the requirement.
    """

    def decorator(func: F) -> F:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            user = get_current_user()
            if user is None:
                raise CustomHTTPException(
                    status_code=401,
                    reason="Authentication required",
                )

            user_roles = set(getattr(user, "roles", []) or [])
            required = set(roles)

            if required:
                if any_of and not required.intersection(user_roles):
                    raise CustomHTTPException(
                        status_code=403,
                        reason="Insufficient role",
                    )
                if not any_of and not required.issubset(user_roles):
                    raise CustomHTTPException(
                        status_code=403,
                        reason="Insufficient role",
                    )

            return func(*args, **kwargs)

        if not hasattr(wrapper, "_decorators"):
            wrapper._decorators = []  # type: ignore[attr-defined]
        decorator.__name__ = "require_roles"
        decorator._args = roles  # type: ignore[attr-defined]
        decorator._any_of = any_of  # type: ignore[attr-defined]
        wrapper._decorators.append(decorator)  # type: ignore[attr-defined]

        return cast(F, wrapper)

    return decorator


def roles_required(*roles: str) -> Callable[[F], F]:
    """Backward compatible wrapper requiring all listed roles.

    Why/How:
        Alias for :func:`require_roles` configured with ``any_of=False`` for
        readability when all roles are mandatory.
    """

    return require_roles(*roles)


def roles_accepted(*roles: str) -> Callable[[F], F]:
    """Backward compatible wrapper requiring any of the listed roles.

    Why/How:
        Alias for :func:`require_roles` configured with ``any_of=True``.
    """

    return require_roles(*roles, any_of=True)


__all__ = ["require_roles", "roles_required", "roles_accepted"]



"""Minimal Flask application illustrating field validators.

The application demonstrates how flarchitect's automatic field validation
works with SQLAlchemy models.  Each column defines validation rules via
the ``info`` mapping which ``flarchitect`` interprets to attach the
appropriate Marshmallow validators.
"""

from __future__ import annotations

from typing import Any

from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import Integer, String
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from flarchitect import Architect

# ``SQLAlchemy`` instance used by the demo app.  ``flarchitect`` will read
# models from this database when initialised inside :func:`create_app`.


class BaseModel(DeclarativeBase):
    """Base model providing a session accessor for flarchitect."""

    def get_session(*args: Any, **kwargs: Any):
        """Return the current database session."""

        return db.session


# create the SQLAlchemy extension using our BaseModel
db = SQLAlchemy(model_class=BaseModel)


class User(db.Model):
    """Simple user model with multiple validated fields.

    Attributes:
        id: Primary key for the user.
        email: User's email address. ``info={'validate': 'email'}`` triggers
            the built-in email validator which rejects invalid addresses.
        homepage: Personal website URL. ``info={'format': 'uri'}`` attaches a
            URL validator ensuring the string is a valid absolute URI.
        slug: Public identifier restricted to URL-friendly characters via the
            ``slug`` validator.
    """

    __tablename__ = "users"

    class Meta:
        """Expose the model through the API and group it under ``User``."""

        tag = "User"
        tag_group = "Demo"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    email: Mapped[str] = mapped_column(
        String(120),
        info={"validate": "email"},  # Rejects values that are not emails
        nullable=False,
    )
    homepage: Mapped[str] = mapped_column(
        String(200),
        info={"format": "uri"},  # Adds URL validation automatically
        nullable=False,
    )
    slug: Mapped[str] = mapped_column(
        String(50),
        info={"validate": "slug"},  # Ensures value is a URL-friendly slug
        nullable=False,
    )


def create_app(config: dict[str, Any] | None = None) -> Flask:
    """Build and configure the validator demo application.

    Args:
        config: Optional mapping to override default configuration values.

    Returns:
        A Flask application exposing REST endpoints for the :class:`User`
        model. Invalid field values raise ``ValidationError`` and return a
        ``400`` response.
    """

    app = Flask(__name__)
    app.config.update(
        SQLALCHEMY_DATABASE_URI="sqlite:///:memory:",
        API_TITLE="Validator API",
        API_VERSION="1.0",
        API_BASE_MODEL=db.Model,
    )
    if config:
        app.config.update(config)

    db.init_app(app)

    with app.app_context():
        db.create_all()
        # ``Architect`` scans the models and registers CRUD routes.
        Architect(app)

    return app


if __name__ == "__main__":  # pragma: no cover - manual execution helper
    create_app().run(debug=True)



"""Tests for JWT hardening: leeway, aud/iss, allowed algs, RS256."""

from __future__ import annotations

import datetime
from collections.abc import Generator

import jwt
import pytest
from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy.pool import StaticPool

from flarchitect.authentication.jwt import decode_token, generate_access_token
from flarchitect.exceptions import CustomHTTPException


db = SQLAlchemy()


class User(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String, unique=True)


@pytest.fixture()
def app_ctx() -> Generator[Flask, None, None]:
    app = Flask(__name__)
    with app.app_context():
        yield app


def test_leeway_allows_small_skew(app_ctx: Flask) -> None:
    """Expired tokens within configured leeway decode successfully."""
    app_ctx.config["API_JWT_LEEWAY"] = 30  # seconds
    secret = "access"
    now = datetime.datetime.now(datetime.timezone.utc)
    payload = {"sub": "user", "iat": now - datetime.timedelta(seconds=20), "exp": now - datetime.timedelta(seconds=10)}
    token = jwt.encode(payload, secret, algorithm="HS256")
    decoded = decode_token(token, secret)
    assert decoded["sub"] == "user"


def test_issuer_audience_enforced_generation_and_decode(app_ctx: Flask) -> None:
    """Tokens include and enforce iss/aud when configured."""
    # Minimal app and DB to use generator
    app_ctx.config.update(
        SQLALCHEMY_DATABASE_URI="sqlite:///:memory:",
        SQLALCHEMY_ENGINE_OPTIONS={"poolclass": StaticPool},
        SQLALCHEMY_TRACK_MODIFICATIONS=False,
        API_USER_MODEL=User,
        API_USER_LOOKUP_FIELD="username",
        ACCESS_SECRET_KEY="access",
    )
    app = app_ctx
    db.init_app(app)
    with app.app_context():
        db.create_all()
        user = User(username="alice")
        db.session.add(user)
        db.session.commit()
        app.config["API_JWT_ISSUER"] = "https://issuer.example"
        app.config["API_JWT_AUDIENCE"] = "api-audience"
        token = generate_access_token(user)
        decoded = decode_token(token, app.config["ACCESS_SECRET_KEY"])
        assert decoded["iss"] == "https://issuer.example"
        assert decoded["aud"] == "api-audience"


def test_wrong_audience_rejected(app_ctx: Flask) -> None:
    app_ctx.config["API_JWT_AUDIENCE"] = "right-aud"
    payload = {
        "sub": "user",
        "iat": datetime.datetime.now(datetime.timezone.utc),
        "exp": datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(minutes=5),
        "aud": "wrong-aud",
    }
    token = jwt.encode(payload, "access", algorithm="HS256")
    with pytest.raises(CustomHTTPException) as exc_info:
        decode_token(token, "access")
    assert exc_info.value.status_code == 401
    assert exc_info.value.reason == "Invalid token"


def test_rs256_generate_and_decode() -> None:
    """Support RS256 with separate private/public keys."""
    try:
        from cryptography.hazmat.primitives import serialization
        from cryptography.hazmat.primitives.asymmetric import rsa
    except Exception:  # pragma: no cover - cryptography not available
        pytest.skip("cryptography not available for RS256 test")

    app = Flask(__name__)
    app.config.update(
        SQLALCHEMY_DATABASE_URI="sqlite:///:memory:",
        SQLALCHEMY_ENGINE_OPTIONS={"poolclass": StaticPool},
        SQLALCHEMY_TRACK_MODIFICATIONS=False,
        API_USER_MODEL=User,
        API_USER_LOOKUP_FIELD="username",
        API_JWT_ALGORITHM="RS256",
    )

    # Generate an RSA keypair
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
    public_key = private_key.public_key()
    private_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    ).decode("utf-8")
    public_pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    ).decode("utf-8")

    app.config["ACCESS_PRIVATE_KEY"] = private_pem
    app.config["ACCESS_PUBLIC_KEY"] = public_pem

    db.init_app(app)
    with app.app_context():
        db.create_all()
        user = User(username="bob")
        db.session.add(user)
        db.session.commit()
        token = generate_access_token(user)
        header = jwt.get_unverified_header(token)
        assert header["alg"] == "RS256"
        decoded = decode_token(token, app.config["ACCESS_PUBLIC_KEY"])  # verify with public key
        assert decoded["username"] == "bob"




from typing import Any

from flask import current_app, g, has_app_context, has_request_context, request
from marshmallow import Schema
from sqlalchemy.orm import DeclarativeBase


def get_config_or_model_meta(
    key: str,
    model: DeclarativeBase | None = None,
    output_schema: Schema | None = None,
    input_schema: Schema | None = None,
    default: Any = None,
    allow_join: bool = False,
    method: str = "IGNORE",
    return_from_config: bool = False,
) -> Any:
    """
    Retrieves configuration or model metadata, prioritizing model metadata.

    Args:
        key (str): The key to search for in the configuration or model meta.
        model (Optional[DeclarativeBase], optional): The SQLAlchemy model.
        output_schema (Optional[Schema], optional): The Marshmallow schema for output.
        input_schema (Optional[Schema], optional): The Marshmallow schema for input.
        default (Any, optional): The default value to return if the key is not found.
        allow_join (bool, optional): Whether to allow joining of results if multiple found.
        method (str, optional): The HTTP method (e.g., 'get', 'post').
        return_from_config (bool, optional): Whether to return the object name from which the value was retrieved.

    Returns:
        Any: The value from the config or model meta, or the default value.
    """

    # from_ob = None

    def normalize_key(key: str) -> str:
        return key.lower()

    def generate_method_based_keys(base_key: str) -> list[str]:
        methods = ["get", "post", "put", "patch", "delete"]
        base_key_lower = base_key.lower()  # Pre-lowercase if needed
        return [f"{meth}_{base_key_lower}" for meth in methods if method.lower() == meth]

    def search_in_sources(sources: list[Any], keys: list[str]) -> Any | None:
        out = []
        for source in sources:
            if source is not None:
                meta = getattr(source, "Meta", None)
                if meta is not None:
                    for key in keys:
                        result = getattr(meta, key, None)
                        if isinstance(result, list) and allow_join:
                            out.extend(result)
                        if result is not None:
                            return result
        if allow_join:
            return out
        return None

    def search_in_flask_config(keys: list[str]) -> Any | None:
        try:
            app = current_app
        except Exception:
            # Outside of an application context; cannot read config here
            return None

        def _lookup() -> Any | None:
            for key in keys:
                upper_key = key.upper()
                prefixed_key = f"API_{upper_key}"

                if upper_key in app.config:
                    return app.config[upper_key]
                if prefixed_key in app.config:
                    return app.config[prefixed_key]

            return None

        if has_app_context():  # Avoid pushing a new context unnecessarily
            return _lookup()

        with app.app_context():  # Ensure config is accessible outside request contexts
            return _lookup()

    normalized_key = normalize_key(key)
    method_based_keys = generate_method_based_keys(normalized_key.replace("api_", ""))

    sources = [model, output_schema, input_schema]
    keys_for_sources = method_based_keys + [
        normalized_key,
        normalize_key(key).replace("api_", ""),
    ]
    keys_for_config = method_based_keys + [normalized_key]

    # Request-local memoization cache (only caches positive lookups)
    cache_key: tuple = (
        normalized_key,
        getattr(model, "__name__", None) if model is not None else None,
        getattr(type(output_schema), "__name__", None) if output_schema is not None else None,
        getattr(type(input_schema), "__name__", None) if input_schema is not None else None,
        bool(allow_join),
        method.lower() if isinstance(method, str) else str(method),
        bool(return_from_config),
    )

    if has_request_context():
        cache = getattr(g, "_flarch_cfg_cache", None)
        if cache is None:
            cache = {}
            g._flarch_cfg_cache = cache
        if cache_key in cache:
            return cache[cache_key]

    # Perform lookups
    model_result = search_in_sources(sources, keys_for_sources)
    if model_result is not None and model_result != [] and model_result != {}:
        out = (model_result, "model") if return_from_config else model_result
        if has_request_context():
            g._flarch_cfg_cache[cache_key] = out
        return out

    cfg_result = search_in_flask_config(keys_for_config)
    if cfg_result is not None and cfg_result != [] and cfg_result != {}:
        out = (cfg_result, "config") if return_from_config else cfg_result
        if has_request_context():
            g._flarch_cfg_cache[cache_key] = out
        return out

    return (default, "default") if return_from_config else default


def is_xml() -> bool:
    """Check if the request is for XML data.

    Returns:
        bool: True if the request is for XML, otherwise False.
    """
    accept_header = request.headers.get("Accept", "")
    content_type_header = request.headers.get("Content-Type", "")
    return any(header in ["application/xml", "text/xml"] for header in [accept_header, content_type_header])



from __future__ import annotations

from collections.abc import Callable
from functools import wraps
from typing import TYPE_CHECKING, Any
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

from flask import Response, request
from sqlalchemy.exc import ProgrammingError
from werkzeug.exceptions import HTTPException

from flarchitect.exceptions import CustomHTTPException
from flarchitect.exceptions import _handle_exception as _handle_exception
from flarchitect.schemas.utils import deserialise_data
from flarchitect.utils.config_helpers import get_config_or_model_meta
from flarchitect.utils.core_utils import convert_case
from flarchitect.utils.general import HTTP_BAD_REQUEST, HTTP_INTERNAL_SERVER_ERROR
from flarchitect.utils.response_helpers import create_response
from flarchitect.utils.responses import serialise_output_with_mallow

if TYPE_CHECKING:  # pragma: no cover - used only for type checking
    from flarchitect.schemas.bases import AutoSchema


def add_dict_to_query(f: Callable) -> Callable:
    """Add a ``dictionary`` key to outputs containing SQLAlchemy row objects.

    Why/How:
        Custom queries often return Core row objects rather than ORM models.
        This wrapper converts those rows to plain dictionaries under
        ``output["dictionary"]`` to make serialisation and testing easier.

    Returns:
        Decorated function that augments the result payload when applicable.
    """

    @wraps(f)
    def decorated_function(*args, **kwargs):
        output = f(*args, **kwargs)
        if isinstance(output, dict):
            try:
                if isinstance(output["query"], list):
                    output["dictionary"] = [result._asdict() for result in output["query"]]
                else:
                    output["dictionary"] = output["query"]._asdict()
            except AttributeError:
                pass
        return output

    return decorated_function


def add_page_totals_and_urls(f: Callable) -> Callable:
    """Attach pagination totals and navigation URLs to a result payload.

    Why/How:
        Uses the current request URL and the returned ``page``, ``limit`` and
        ``total_count`` values to calculate ``next_url``, ``previous_url``,
        ``current_page`` and ``total_pages``.

    Args:
        f: Function to decorate.

    Returns:
        Decorated function with pagination keys added when possible.
    """

    @wraps(f)
    def decorated_function(*args, **kwargs):
        output = f(*args, **kwargs)
        limit, page, total_count = (
            output.get("limit"),
            output.get("page"),
            output.get("total_count"),
        )

        next_url, previous_url, current_page, total_pages = None, None, None, None

        if total_count and limit:
            total_pages = -(-total_count // limit)  # Ceiling division
            current_page = page

            parsed_url = urlparse(request.url)
            query_params = parse_qs(parsed_url.query)

            query_params["limit"] = [str(limit)]
            next_page, prev_page = page + 1, page - 1
            next_url = _construct_url(parsed_url, query_params, next_page, total_count, limit)
            previous_url = _construct_url(parsed_url, query_params, prev_page, total_count, limit)

        if isinstance(output, dict):
            output.update(
                {
                    "next_url": next_url,
                    "previous_url": previous_url,
                    "current_page": current_page,
                    "total_pages": total_pages,
                }
            )

        return output

    return decorated_function


def _construct_url(parsed_url, query_params, page, total_count, limit):
    """Construct a pagination URL for a specific page if within bounds.

    Args:
        parsed_url: Parsed URL tuple from :func:`urllib.parse.urlparse`.
        query_params: Mapping of query parameters to encode.
        page: Target page number.
        total_count: Total number of items.
        limit: Items per page.

    Returns:
        The constructed URL string or ``None`` if out of range.
    """
    if 0 < page <= total_count // limit:
        query_params["page"] = [str(page)]
        return urlunparse(parsed_url._replace(query=urlencode(query_params, doseq=True)))
    return None


def handle_many(output_schema: type[AutoSchema], input_schema: type[AutoSchema] | None = None) -> Callable:
    """Serialise a list response and optionally deserialise input.

    Why/How:
        Wraps a route that returns multiple items, applying the supplied
        Marshmallow schema for output and, if provided, validating input via
        ``input_schema``.

    Args:
        output_schema: Marshmallow schema class used to serialise the output.
        input_schema: Optional schema to deserialise request data.

    Returns:
        The decorated function.
    """
    return _handle_decorator(output_schema, input_schema, many=True)


def handle_one(output_schema: type[AutoSchema], input_schema: type[AutoSchema] | None = None) -> Callable:
    """Serialise a single-item response and optionally deserialise input.

    Args:
        output_schema: Marshmallow schema class used to serialise the output.
        input_schema: Optional schema to deserialise request data.

    Returns:
        The decorated function.
    """
    return _handle_decorator(output_schema, input_schema, many=False)


def _handle_decorator(
    output_schema: type[AutoSchema],
    input_schema: type[AutoSchema] | None,
    many: bool,
) -> Callable:
    """Base decorator to handle input and output using Marshmallow schemas.

    Args:
        output_schema: Schema used to serialise output.
        input_schema: Schema used to deserialise input.
        many: Whether the route returns multiple records.

    Returns:
        The decorated function.
    """

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        @standardise_response
        @fields(output_schema, many=many)
        def wrapper(*args: Any, **kwargs: dict[str, Any]) -> dict[str, Any] | tuple:
            if input_schema:
                data_or_error = deserialise_data(input_schema, request)
                if isinstance(data_or_error, tuple):  # Error occurred during deserialization
                    case = get_config_or_model_meta("API_FIELD_CASE", default="snake")
                    error = {convert_case(k, case): v for k, v in data_or_error[0].items()}
                    raise CustomHTTPException(HTTP_BAD_REQUEST, error)
                kwargs["deserialized_data"] = data_or_error
                kwargs["model"] = getattr(input_schema.Meta, "model", None)

            new_output_schema: type[AutoSchema] | None = kwargs.pop("schema", None)
            result = func(*args, **kwargs)

            return serialise_output_with_mallow(new_output_schema, result) if new_output_schema else result

        return wrapper

    return decorator


def standardise_response(func: Callable) -> Callable:
    """Standardise API responses and trigger error callbacks consistently.

    Why/How:
        Ensures responses follow a uniform envelope via
        :func:`flarchitect.utils.response_helpers.create_response`, catches
        common exceptions and routes them through a central handler, then calls
        any configured error callback.

    Args:
        func: The route handler to wrap.

    Returns:
        A wrapper that returns a :class:`flask.Response`.
    """

    @wraps(func)
    def decorated_function(*args: Any, **kwargs: Any) -> Response:
        had_error = False
        error: Any | None = None
        status_code: int | None = None
        value: Any | None = None

        try:
            result = func(*args, **kwargs)
            out_resp = create_response(result=result)
            status_code = out_resp.status_code
            payload = out_resp.get_json(silent=True) or {}
            value = payload.get("value")
            error = payload.get("errors")
            if error or status_code > 299:
                had_error = True

        except HTTPException as e:
            had_error = True
            error = e.description
            status_code = e.code or HTTP_INTERNAL_SERVER_ERROR
            value = {"error": e.name, "reason": e.description}
            out_resp = _handle_exception(error, status_code, e.name, print_exc=True)

        except ProgrammingError as e:
            had_error = True
            text = str(e).split(")")[1].split("\n")[0].strip().capitalize()
            error = f"SQL Format Error: {text}"
            status_code = HTTP_BAD_REQUEST
            value = None
            out_resp = _handle_exception(error, status_code)

        except CustomHTTPException as e:
            had_error = True
            error = e.reason
            status_code = e.status_code
            value = None
            out_resp = _handle_exception(error, status_code, e.error)

        except Exception as e:
            had_error = True
            error = str(e)
            status_code = HTTP_INTERNAL_SERVER_ERROR
            value = None
            out_resp = _handle_exception(f"Internal Server Error: {error}", status_code)

        finally:
            error_callback = get_config_or_model_meta("API_ERROR_CALLBACK")
            if error_callback and had_error:
                error_callback(error, status_code, value)

        return out_resp

    return decorated_function


# Backwards-compatible alias (US spelling)
def standardize_response(func: Callable) -> Callable:  # pragma: no cover - shim
    return standardise_response(func)


def fields(model_schema: type[AutoSchema], many: bool = False) -> Callable:
    """Control which fields are serialised on the response.

    Why/How:
        Reads an optional ``fields`` query parameter and re‑instantiates the
        output schema with ``only`` to reduce payload size. Falls back to the
        full schema when the parameter is absent.

    Args:
        model_schema: Marshmallow schema class for the response.
        many: Whether the endpoint returns a collection.

    Returns:
        The decorated function.
    """

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: dict[str, Any]) -> Any:
            if request.method == "DELETE":
                return func(*args, **kwargs)

            select_fields = request.args.get("fields")
            if select_fields and get_config_or_model_meta("API_ALLOW_SELECT_FIELDS", model_schema.Meta.model, default=True):
                select_fields = [field.split(".")[-1] for field in select_fields.split(",")]
                if callable(model_schema):
                    kwargs["schema"] = model_schema(many=many, only=select_fields)
                else:
                    kwargs["schema"] = model_schema.__class__(many=many, only=select_fields)
            else:
                if callable(model_schema):
                    kwargs["schema"] = model_schema(many=many)
                else:
                    kwargs["schema"] = model_schema.__class__(many=many)

            return func(*args, **kwargs)

        return wrapper

    return decorator



"""Tests for role-based access decorators."""

from collections.abc import Generator
from types import SimpleNamespace

import pytest
from flask import Flask
from flask.testing import FlaskClient
from marshmallow import Schema, fields

from flarchitect import Architect
from flarchitect.authentication import require_roles
from flarchitect.authentication.user import set_current_user
from flarchitect.exceptions import CustomHTTPException
from flarchitect.specs.utils import handle_authorization
from flarchitect.utils.response_helpers import create_response


def test_require_roles_all() -> None:
    """The decorator blocks access when required roles are missing."""

    @require_roles("admin")
    def sample() -> str:
        return "ok"

    set_current_user(SimpleNamespace(roles=["admin"]))
    assert sample() == "ok"

    set_current_user(None)
    with pytest.raises(CustomHTTPException) as exc:
        sample()
    assert exc.value.status_code == 401
    assert exc.value.reason == "Authentication required"

    set_current_user(SimpleNamespace(roles=["user"]))
    with pytest.raises(CustomHTTPException) as exc:
        sample()
    assert exc.value.status_code == 403
    assert exc.value.reason == "Insufficient role"


@pytest.fixture()
def client_roles() -> Generator[FlaskClient, None, None]:
    """Create a Flask client with a route enforcing roles."""

    app = Flask(__name__)
    app.config.update(FULL_AUTO=False, API_CREATE_DOCS=False)

    with app.app_context():
        architect = Architect(app=app)
        holder = SimpleNamespace(user=None)

        class PingSchema(Schema):
            status = fields.Str()

        @app.before_request
        def load_user() -> None:
            set_current_user(holder.user)

        @app.errorhandler(CustomHTTPException)
        def handle_custom(exc: CustomHTTPException):
            return create_response(
                status=exc.status_code,
                errors={"error": exc.error, "reason": exc.reason},
            )

        @app.route("/protected")
        @architect.schema_constructor(output_schema=PingSchema, roles=("admin",))
        def protected() -> dict[str, str]:
            return {"status": "ok"}

        client = app.test_client()
        yield client, holder


def test_schema_constructor_applies_roles(
    client_roles: tuple[FlaskClient, SimpleNamespace],
) -> None:
    """Access is restricted based on user roles when configured."""

    client, holder = client_roles

    holder.user = SimpleNamespace(roles=["admin"])
    resp = client.get("/protected")
    assert resp.status_code == 200

    holder.user = None
    resp = client.get("/protected")
    assert resp.status_code == 401

    holder.user = SimpleNamespace(roles=["user"])
    resp = client.get("/protected")
    assert resp.status_code == 403


def test_openapi_documents_roles() -> None:
    """OpenAPI specification documents required roles."""

    spec_template = {"parameters": [], "responses": {"401": {"description": ""}}}

    @require_roles("admin", "editor")
    def view() -> None:  # pragma: no cover - simple callable
        pass

    handle_authorization(view, spec_template)

    assert spec_template["security"] == [{"bearerAuth": []}]
    assert "Roles required: admin, editor" in spec_template["responses"]["401"]["description"]


def test_require_roles_any() -> None:
    """Access granted when user has any matching role."""

    @require_roles("admin", "editor", any_of=True)
    def sample() -> str:
        return "ok"

    set_current_user(SimpleNamespace(roles=["editor"]))
    assert sample() == "ok"

    set_current_user(None)
    with pytest.raises(CustomHTTPException) as exc:
        sample()
    assert exc.value.status_code == 401
    assert exc.value.reason == "Authentication required"

    set_current_user(SimpleNamespace(roles=["user"]))
    with pytest.raises(CustomHTTPException) as exc:
        sample()
    assert exc.value.status_code == 403
    assert exc.value.reason == "Insufficient role"


def test_openapi_documents_roles_accepted() -> None:
    """OpenAPI specification documents accepted roles."""

    spec_template = {"parameters": [], "responses": {"401": {"description": ""}}}

    @require_roles("admin", "editor", any_of=True)
    def view() -> None:  # pragma: no cover - simple callable
        pass

    handle_authorization(view, spec_template)

    assert spec_template["security"] == [{"bearerAuth": []}]
    assert "Roles accepted: admin, editor" in spec_template["responses"]["401"]["description"]



from __future__ import annotations

import graphene
from sqlalchemy import Integer
from sqlalchemy.orm import DeclarativeBase, Mapped, Session, mapped_column
from sqlalchemy.types import TypeDecorator
from sqlalchemy import create_engine

from flarchitect.graphql import _convert_sqla_type, create_schema_from_models


class Base(DeclarativeBase):
    pass


class Unknown(TypeDecorator):
    impl = Integer


def test_convert_sqla_type_fallback_to_string() -> None:
    # Unknown (unmapped) types should map to graphene.String
    assert _convert_sqla_type(Unknown(), {}) is graphene.String


class Foo(Base):
    __tablename__ = "foo"
    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    name: Mapped[str] = mapped_column(Integer)


def test_update_delete_when_missing_row() -> None:
    engine = create_engine("sqlite:///:memory:")
    Base.metadata.create_all(engine)
    session = Session(engine)
    schema = create_schema_from_models([Foo], session)

    # Update non-existent returns null
    result = schema.execute("mutation { update_foo(id: 999, name: 1) { id } }")
    assert result.errors is None
    assert result.data == {"update_foo": None}

    # Delete non-existent returns false
    result = schema.execute("mutation { delete_foo(id: 999) }")
    assert result.errors is None
    assert result.data == {"delete_foo": False}




"""Tests for ``get_html_path`` helper."""

import importlib.util
import sys
import types
from pathlib import Path

from flask import Flask

# Create minimal stubs so ``general`` can be imported without the full package.
sys.modules.setdefault("flarchitect", types.ModuleType("flarchitect"))
utils_mod = types.ModuleType("flarchitect.utils")
config_helpers = types.ModuleType("flarchitect.utils.config_helpers")
core_utils = types.ModuleType("flarchitect.utils.core_utils")
config_helpers.get_config_or_model_meta = lambda *args, **kwargs: None
core_utils.convert_case = lambda value, case=None: value
core_utils.get_count = lambda *args, **kwargs: 0
utils_mod.config_helpers = config_helpers
utils_mod.core_utils = core_utils
sys.modules["flarchitect.utils"] = utils_mod
sys.modules["flarchitect.utils.config_helpers"] = config_helpers
sys.modules["flarchitect.utils.core_utils"] = core_utils

_general_spec = importlib.util.spec_from_file_location(
    "general",
    Path(__file__).resolve().parents[1] / "flarchitect" / "utils" / "general.py",
)
assert _general_spec and _general_spec.loader  # for mypy
_general = importlib.util.module_from_spec(_general_spec)
_general_spec.loader.exec_module(_general)
sys.modules["flarchitect.utils.general"] = _general

# Clean up stubbed modules to avoid polluting subsequent tests that import
# the real package. The loaded general module remains available via the
# explicit sys.modules entry above.
for _mod in [
    "flarchitect.utils.core_utils",
    "flarchitect.utils.config_helpers",
    "flarchitect.utils",
    "flarchitect",
]:
    sys.modules.pop(_mod, None)

get_html_path = _general.get_html_path


def test_get_html_path_falls_back() -> None:
    """``get_html_path`` should locate the package ``html`` directory without an app."""
    path = get_html_path()
    assert path.endswith("html")
    assert (Path(path) / "apispec.html").exists()


def test_get_html_path_uses_extension() -> None:
    """``get_html_path`` should use the registered extension when available."""
    app = Flask(__name__)

    class DummyExt:
        def get_templates_path(self) -> str:
            return "/tmp/custom_html"

    app.extensions["flarchitect"] = DummyExt()
    with app.app_context():
        assert get_html_path() == "/tmp/custom_html"
